from typing import NoReturn, Tuple, List
import copy
import numpy
from fractions import Fraction

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary
from cryspy.A_functions_base.function_1_strings import \
    transform_string_to_r_b, transform_r_b_to_string
from cryspy.A_functions_base.function_2_space_group import \
    get_shift_by_centring_type, mult_matrix_vector, mult_matrixes

from cryspy.A_functions_base.function_2_sym_elems import \
    form_symm_elems_by_b_i_r_ij

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class SpaceGroupSymop(ItemN):
    """Symmetry space group operation.
    
    Contains information about the symmetry operations of the
    space group.

    Attributes
    ----------
        - operation_xyz (mandatory)
        - id, operation_description, generator_xyz, sg_id (optional)
        - r, b, r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33, b_1, b_2,
          b_3 (internal attributes)

    Methods
    -------------
        - define_by_el_symm() (class method)
        - get_symop_inversed()
        - get_symops_by_centring_type()
        - get_symops_by_generator_xyz()
        - get_coords_xyz_by_coord_xyz()
        
    """
    ATTR_MANDATORY_NAMES = ("operation_xyz", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("operation_xyz", )

    ATTR_OPTIONAL_NAMES = ("id", "operation_description", "generator_xyz",
                           "sg_id")
    ATTR_OPTIONAL_TYPES = (str, str, str, int)
    ATTR_OPTIONAL_CIF = ("id", "operation_description", "generator_xyz",
                         "sg_id")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("r", "b", "r_11", "r_12", "r_13", "r_21", "r_22",
                      "r_23", "r_31", "r_32", "r_33", "b_1", "b_2", "b_3")
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "space_group_symop"

    def __init__(self, **kwargs) -> NoReturn:
        super(SpaceGroupSymop, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self) -> NoReturn:
        operation_xyz = self.operation_xyz
        if operation_xyz is None:
            return False
        r, b = transform_string_to_r_b(operation_xyz, labels=("x", "y", "z"))

        self.__dict__["r"] = r
        self.__dict__["b"] = b
        self.__dict__["r_11"] = r[0, 0]
        self.__dict__["r_12"] = r[0, 1]
        self.__dict__["r_13"] = r[0, 2]
        self.__dict__["r_21"] = r[1, 0]
        self.__dict__["r_22"] = r[1, 1]
        self.__dict__["r_23"] = r[1, 2]
        self.__dict__["r_31"] = r[2, 0]
        self.__dict__["r_32"] = r[2, 1]
        self.__dict__["r_33"] = r[2, 2]
        self.__dict__["b_1"] = b[0]
        self.__dict__["b_2"] = b[1]
        self.__dict__["b_3"] = b[2]

    @classmethod
    def define_by_el_symm(cls, el_symm):
        """Define object by element of symmetry (class method).
        
        Arguments
        ---------
            - el_symm is [b_1, r_11, r_12, r_13,
                          b_2, r_21, r_22, r_23,
                          b_3, r_31, r_32, r_33]
        """
        [b_1, r_11, r_12, r_13, b_2, r_21, r_22, r_23, b_3, r_31, r_32, r_33] \
            = el_symm
        r = numpy.array([[r_11, r_12, r_13], [r_21, r_22, r_23],
                         [r_31, r_32, r_33]], dtype=int)
        b = numpy.array([b_1, b_2, b_3], dtype=float)
        symop = transform_r_b_to_string(r, b, labels=("x", "y", "z"))
        obj = cls(operation_xyz=symop)
        return obj

    def get_symop_inversed(self, pcentr=numpy.zeros(3, float)):
        r, b = self.r, self.b
        r_new = -1*r
        b_new = -1*b+2*pcentr
        _symop = transform_r_b_to_string(r_new, b_new, labels=("x", "y", "z"))
        _item = SpaceGroupSymop(operation_xyz=_symop)
        return _item

    def get_symops_by_centring_type(self, centring_type:str)->Tuple[str]:
        r, b = self.r, self.b
        r_new = r
        shift = get_shift_by_centring_type(centring_type)
        symops = []
        for _shift in shift:
            b_new =  b + numpy.array(_shift, dtype=Fraction)
            _symop = transform_r_b_to_string(r_new, b_new, labels=("x", "y", "z"))
            symop = SpaceGroupSymop(operation_xyz=_symop)
            symops.append(symop)
        return symops

    def get_symops_by_generator_xyz(self, generator_xyz:str)->Tuple[str]:
        r, b = self.r, self.b
        symop = []
        W_g, w_g = transform_string_to_r_b(generator_xyz, labels=("x", "y", "z"))
        
        flag = True
        W_i = copy.deepcopy(r)
        w_i = copy.deepcopy(b)%1
        _symop = transform_r_b_to_string(W_i, w_i, labels=("x", "y", "z"))
        symop.append(_symop)
        i_stop = 1
        while flag:
            i_stop += 1
            w_i = (mult_matrix_vector(W_g, w_i) + w_g)%1 # w3 = W2 x w1 + w2
            W_i = mult_matrixes(W_g, W_i)          # W3 = W2 x W1
            _symop = transform_r_b_to_string(W_i, w_i, labels=("x", "y", "z"))
            if ((_symop in symop) | (i_stop > 100)):
                flag = False
            else:
                symop.append(_symop)
        res = []
        gen_orig = ""
        if not(self.generator_xyz is None): gen_orig = f"{self.generator_xyz:}"
        res =[SpaceGroupSymop(id=f"{_i+1:}", operation_xyz=_symop, generator_xyz=f"{gen_orig:}_{generator_xyz:}") for _i, _symop in enumerate(symop)]
        return res

    def get_coords_xyz_by_coord_xyz(self, coord_xyz: str) -> Tuple:
        _item = SpaceGroupSymop(operation_xyz=coord_xyz)
        generator_xyz = self.operation_xyz
        res = [_.operation_xyz for _ in _item.get_symops_by_generator_xyz(generator_xyz)]
        return tuple(res)



class SpaceGroupSymopL(LoopN):
    """Symmetry space group operations.

    Contains information about the symmetry operations of the space group.
    """
    ITEM_CLASS = SpaceGroupSymop
    ATTR_INDEX = "id"
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(SpaceGroupSymopL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    @classmethod
    def create_by_generators_xyz(cls, generators_xyz: List):
        """
        Create list of symmetry operators by generators.

        Parameters
        ----------
        cls : TYPE
            DESCRIPTION.
        generators_xyz : List
            DESCRIPTION.

        Returns
        -------
        _obj : TYPE
            DESCRIPTION.

        """
        _item = SpaceGroupSymop(operation_xyz="x,y,z")
        items_1 = [_item]
        for generator_xyz in generators_xyz:
            items_2 = []
            for _item in items_1:
                items_2.extend(_item.get_symops_by_generator_xyz(
                    generator_xyz))
            items_1 = items_2
        symop = []
        item_uniq = []
        _i = 0
        for _item in items_1:
            if not(_item.operation_xyz in symop):
                _i += 1
                symop.append(_item.operation_xyz)
                _item.id = f"{_i:}"
                item_uniq.append(_item)
        _obj = cls()
        _obj.items = item_uniq
        return _obj

    def get_coords_xyz_by_coord_xyz(self, coord_xyz: str) -> Tuple:
        """
        Calculate new fraction for given fraction.

        Parameters
        ----------
        coord_xyz : str
            DESCRIPTION.

        Returns
        -------
        Tuple
            DESCRIPTION.

        """
        items = self.items
        res = []
        for _item in items:
            res.extend(_item.get_coords_xyz_by_coord_xyz(coord_xyz))
        return tuple(frozenset(res))

    def get_symop_for_x1_x2(self, fract_xyz_1: numpy.ndarray,
                            fract_xyz_2: numpy.ndarray):
        """
        Calculate symmetry operators transforming one fraction to another.

        Parameters
        ----------
        fract_xyz_1 : numpy.ndarray
            DESCRIPTION.
        fract_xyz_2 : numpy.ndarray
            DESCRIPTION.

        Returns
        -------
        res : TYPE
            DESCRIPTION.

        """
        np_fract_xyz_1 = numpy.array(fract_xyz_1, dtype=float)
        np_fract_xyz_2 = numpy.array(fract_xyz_2, dtype=float)
        res = SpaceGroupSymopL()
        l_item = []
        for item_s_g_s in self.items:
            m_r = numpy.array(item_s_g_s.r, dtype=float)
            v_b = numpy.array(item_s_g_s.b, dtype=float)
            if numpy.allclose(numpy.mod(numpy.dot(m_r, np_fract_xyz_1)+v_b, 1),
                              numpy.mod(np_fract_xyz_2, 1)):
                l_item.append(item_s_g_s)
        res.items = l_item
        return res

    def get_symm_elems(self):
        r_ij = (self.r_11, self.r_12, self.r_13, self.r_21, self.r_22, self.r_23, self.r_31, self.r_32, self.r_33)
        b_i = (self.b_1, self.b_2, self.b_3)
        symm_elems = form_symm_elems_by_b_i_r_ij(b_i, r_ij)
        return symm_elems
    
# s_cont = """
# loop_
# _space_group_symop.id
# _space_group_symop.operation_xyz
# _space_group_symop.operation_description
#   1    x,y,z              'identity mapping'
#   2    -x,-y,-z           'inversion'
#   3    -x,1/2+y,1/2-z '2-fold screw rotation with axis in (0,y,1/4)'
#   4    x,1/2-y,1/2+z  'c glide reflection through the plane (x,1/4,y)'
# """

# obj = SpaceGroupSymopL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["3"], end="\n\n")
# print(obj["3"].r_11, end="\n\n")
# print(obj.numpy_operation_xyz, end="\n\n")
# print(obj.numpy_r_11.astype(int), end="\n\n")



"""AtomRhoOrbitalRadialSlater and AtomRhoOrbitalRadialSlaterL classes."""

from typing import NoReturn
import math
import numpy
import scipy
import scipy.optimize


from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.A_functions_base.orbital_functions import calc_jl

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.C_item_loop_classes.cl_1_atom_type_scat import AtomTypeScat


class AtomRhoOrbitalRadialSlater(ItemN):
    """Description of radial Slater functions.

    These items are used when the radial dependence of the
    electron density, R(k(l), l, r), of the atom specified in
    _atom_rho_orbital_atom_label is expressed as a Slater-type
    function [Hansen & Coppens (1978), equation (3)]:

    Attributes
        - n0, zeta0 (mandatory)
        - "coeff_1s", "coeff_2s", "coeff_3s", "coeff_4s", "coeff_5s", "coeff_6s",
          "coeff_2p", "coeff_3p", "coeff_4p", "coeff_5p", "coeff_6p",
          "coeff_3d", "coeff_4d", "coeff_5d", "coeff_6d",
          "coeff_4f", "coeff_5f", "coeff_6f" (optional)
    """

    ATTR_MANDATORY_NAMES = ("n0", "zeta0")
    ATTR_MANDATORY_TYPES = (int, float)
    ATTR_MANDATORY_CIF = ("n0", "zeta0")

    ATTR_OPTIONAL_NAMES = (
        "coeff_1s", "coeff_2s", "coeff_3s", "coeff_4s", "coeff_5s", "coeff_6s",
        "coeff_2p", "coeff_3p", "coeff_4p", "coeff_5p", "coeff_6p",
        "coeff_3d", "coeff_4d", "coeff_5d", "coeff_6d",
        "coeff_4f", "coeff_5f", "coeff_6f")
    ATTR_OPTIONAL_TYPES = (
        float, float, float, float, float, float,
        float, float, float, float, float,
        float, float, float, float,
        float, float, float)
    ATTR_OPTIONAL_CIF = (
        "coeff_1s", "coeff_2s", "coeff_3s", "coeff_4s", "coeff_5s", "coeff_6s",
        "coeff_2p", "coeff_3p", "coeff_4p", "coeff_5p", "coeff_6p",
        "coeff_3d", "coeff_4d", "coeff_5d", "coeff_6d",
        "coeff_4f", "coeff_5f", "coeff_6f")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("zeta0",
        "coeff_1s", "coeff_2s", "coeff_3s", "coeff_4s", "coeff_5s", "coeff_6s",
        "coeff_2p", "coeff_3p", "coeff_4p", "coeff_5p", "coeff_6p",
        "coeff_3d", "coeff_4d", "coeff_5d", "coeff_6d",
        "coeff_4f", "coeff_5f", "coeff_6f")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "atom_rho_orbital_radial_slater"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomRhoOrbitalRadialSlater, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"n0": 0, "zeta0": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_normalized_rho(self, radius: numpy.array, kappa: float = 1.) -> \
            numpy.array:
        """Give normalized Slatter type function.

        R = ((2*zeta0)**(n0+0.5)/((2*n0)!)**0.5)*r**(n0-1)*exp(-zeta0*r)
        (see Clementi_1974)

        Arguments
        ---------
            - radius is numpy 1D array of float numbers in Angstrem
            - kappa is describes contraction/extension of atomic orbital

        Output
        ------
            - normalized radial density given as numpy 1D array of floats

        Example
        -------
            >>> radius = numpy.linspace(0.,5,600)
            >>> rho = obj.calc_normalized_rho(radius)
        """
        # transformation from atomic units to inverse angstrems
        zeta0 = self.zeta0
        n0 = self.n0
        zeta_ang = kappa*zeta0/0.52918
        coeff_norm = ((2.*zeta_ang)**(float(n0)+0.5)) / \
            math.sqrt(math.factorial(2*n0))
        norm_density = coeff_norm*(radius**(n0-1))*numpy.exp(-zeta_ang*radius)
        return norm_density

    def calc_jl(self, sthovl: numpy.array, l_max: int, kappa: float = 1.) -> \
            numpy.array:
        """Calculate jl for l from 0 until l_max of atomic orbital."""
        zeta0 = self.zeta0
        n0 = self.n0
        jl = calc_jl(sthovl, [1.], [n0], [zeta0, ], kappa=kappa, l_max=l_max)
        return jl


class AtomRhoOrbitalRadialSlaterL(LoopN):
    """Description of radial Slater functions.

    These items are used when the radial dependence of the
    electron density, R(k(l), l, r), of the atom specified in
    _atom_rho_orbital_atom_label is expressed as a Slater-type
    function [Hansen & Coppens (1978), equation (3)]:
    """

    ITEM_CLASS = AtomRhoOrbitalRadialSlater
    ATTR_INDEX = None # "atom_label"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(AtomRhoOrbitalRadialSlaterL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def calc_normalized_rho(self, radius: numpy.array, shell: str, kappa=1.,
                            *argv) -> numpy.array:
        """Calc normalized rho.

        Give normalized Slatter type function at given radius for given shell.

        $$
        R = sum_i coeff_i*((2*zeta0)**(n0+0.5)/(2*n0)!)*r**(n0-1)*exp(-zeta0*r)
        $$
        (see Clementi_1974)

        Arguments
        ---------
            - radius is numpy 1D array of float numbers in Angstrem
            - shell is it is orbital shell
            - kappa is describes contraction/extension of atomic orbital

        Output
        ------
            - normalized radial density given as numpy 1D array of floats

        Example
        -------
            >>> radius = numpy.linspace(0.,5,600)
            >>> rho = obj.calc_rho(radius, "3d")
            >>> rho = obj.calc_rho(radius, "3d", "2s")
        """
        # transformation from atomic units to inverse angstrems
        l_shell = [shell]
        l_shell.extend(argv)
        n_shell = len(l_shell)
        n_item = len(self.items)
        coeff = numpy.zeros((n_shell, n_item), dtype=float)

        norm_density = numpy.zeros((radius.size, n_shell), dtype=float)

        l_attr = [f"coeff_{_.strip():}" for _ in l_shell]
        for _i, _item in enumerate(self.items):
            normalized_rho = _item.calc_normalized_rho(radius, kappa=kappa)
            for _j, _attr in enumerate(l_attr):
                coeff[_j, _i] = float(getattr(_item, _attr))
            norm_density += normalized_rho[:, numpy.newaxis] * \
                (coeff[:, _i])[numpy.newaxis, :]
        return norm_density


    def calc_jl_by_radial_density(self, sthovl, lmax: int, shell: str,
                                  kappa=1.):
        """Calculate <j_l> for given sthovl, l and shell.

        $$
        <j_l> = sum_{n, dzeta} integral_{0}^{inf} R^{2}(n,dzeta,r) j_l(r*s) dr
        $$
        """
        n0 = numpy.array(self.n0, dtype=int)
        zeta0 = numpy.array(self.zeta0, dtype=float)
        coeff = numpy.array(getattr(self, f"coeff_{shell:}"), dtype=float)

        if any([_ is None for _ in coeff]):
            return None

        jl = calc_jl(sthovl, coeff, n0, zeta0, kappa=kappa, l_max = 3)

        return jl

    def refine_coefficients_by_jl(self, atom_type: str, shell: str,
                                  sthovl_min=0.0, sthovl_max=2.0, lande=2.,
                                  kappa=1., sthvl_num=100):
        """Refine coefficients by jl."""
        ats = AtomTypeScat.form_by_symbol(atom_type)

        sthovl = numpy.linspace(sthovl_min, sthovl_max, num=int(sthvl_num))
        ff = ats.calc_form_factor(sthovl, lande=lande, kappa=kappa)
        fitable = self.get_variables()

        def tempfunc(param):
            for _1, _2 in zip(fitable, param):
                _1.value = abs(_2)  # only positive coefficients
            j0 = self.calc_jl_by_radial_density(sthovl, 0, shell, kappa)[0]
            chi_sq = (numpy.square(ff - j0)).sum()
            return chi_sq

        param_0 = [_.value for _ in fitable]
        j0 = self.calc_jl_by_radial_density(sthovl, 0, shell, kappa)[0]

        res = scipy.optimize.basinhopping(tempfunc, param_0, niter=10,
                                          T=0.1, stepsize=0.1, interval=20,
                                          disp=True)
        for _1, _2 in zip(fitable, res.x):
            _1.value = abs(_2)  # only positive coefficients, found minima

        return res



from typing import Callable
import numpy

na = numpy.newaxis

def calc_m_sq_sin_sq_para(tensor_sigma, flag_tensor_sigma: bool = False):
    """Calculate the term P1 for paramagnetic sublattice.
    For details see documentation "Integrated intensity from powder diffraction".
    """
    sigma_11 = tensor_sigma[0]
    sigma_12 = tensor_sigma[1]
    sigma_22 = tensor_sigma[4]
    
    p_1 = 0.5*(numpy.square(numpy.abs(sigma_11)) + numpy.square(numpy.abs(sigma_22))) + \
        numpy.square(numpy.abs(sigma_12))

    dder = {}
    if flag_tensor_sigma:
        ones = numpy.ones_like(sigma_11.real)
        dder_sigma_11_real = sigma_11.real * ones
        dder_sigma_11_imag = sigma_11.imag * ones
        dder_sigma_22_real = sigma_22.real * ones
        dder_sigma_22_imag = sigma_22.imag * ones
        dder_sigma_12_real = 2*sigma_12.real * ones
        dder_sigma_12_imag = 2*sigma_12.imag * ones
        zeros = numpy.zeros_like(dder_sigma_11_real)
        dder["tensor_sigma_real"] = numpy.stack([
            dder_sigma_11_real, dder_sigma_12_real, zeros,
            zeros, dder_sigma_22_real, zeros, zeros, zeros, zeros], axis=0)
        dder["tensor_sigma_imag"] = numpy.stack([
            dder_sigma_11_imag, dder_sigma_12_imag, zeros,
            zeros, dder_sigma_22_imag, zeros, zeros, zeros, zeros], axis=0)
    return p_1, dder


def calc_m_sq_cos_sq_para(tensor_sigma, flag_tensor_sigma: bool = False):
    """Calculate the term P2 for paramagnetic sublattice.
    For details see documentation "Integrated intensity from powder diffraction".
    """
    sigma_13 = tensor_sigma[2]
    sigma_23 = tensor_sigma[5]
    
    p_2 = numpy.square(numpy.abs(sigma_13)) + numpy.square(numpy.abs(sigma_23))

    dder = {}
    if flag_tensor_sigma:
        ones = numpy.ones_like(sigma_13.real)
        dder_sigma_13_real = 2*sigma_13.real * ones
        dder_sigma_13_imag = 2*sigma_13.imag * ones
        dder_sigma_23_real = 2*sigma_23.real * ones
        dder_sigma_23_imag = 2*sigma_23.imag * ones
        zeros = numpy.zeros_like(dder_sigma_13_real)
        dder["tensor_sigma_real"] = numpy.stack([
            zeros, zeros, dder_sigma_13_real,
            zeros, zeros, dder_sigma_23_real, zeros, zeros, zeros], axis=0)
        dder["tensor_sigma_imag"] = numpy.stack([
            zeros, zeros, dder_sigma_13_imag,
            zeros, zeros, dder_sigma_23_imag, zeros, zeros, zeros], axis=0)
    return p_2, dder


def calc_cross_term_para(f_nucl, tensor_sigma, flag_f_nucl: bool = False, flag_tensor_sigma: bool = False):
    """Calculate the term P3 for paramagnetic sublattice.
    For details see documentation "Integrated intensity from powder diffraction".
    """
    sigma_11 = tensor_sigma[0]
    sigma_22 = tensor_sigma[4]
    
    p_3 = f_nucl.real * (sigma_11.real + sigma_22.real) + f_nucl.imag * (sigma_11.imag + sigma_22.imag)

    dder = {}
    if flag_f_nucl:
        dder["f_nucl_real"] = (sigma_11.real + sigma_22.real)*numpy.ones_like(f_nucl.real)
        dder["f_nucl_imag"] = (sigma_11.imag + sigma_22.imag)*numpy.ones_like(f_nucl.imag)

    if flag_tensor_sigma:
        ones = numpy.ones_like(sigma_11.real)
        dder_sigma_11_real = f_nucl.real * ones
        dder_sigma_11_imag = f_nucl.imag * ones
        dder_sigma_22_real = f_nucl.real * ones
        dder_sigma_22_imag = f_nucl.imag * ones
        zeros = numpy.zeros_like(dder_sigma_11_real)
        dder["tensor_sigma_real"] = numpy.stack([
            dder_sigma_11_real, zeros, zeros, 
            zeros, dder_sigma_22_real, zeros, zeros, zeros, zeros], axis=0)
        dder["tensor_sigma_imag"] = numpy.stack([
            dder_sigma_11_imag, zeros, zeros, 
            zeros, dder_sigma_22_imag, zeros, zeros, zeros, zeros], axis=0)
    return p_3, dder


def calc_chiral_term_cos_sin_sq_para(tensor_sigma, flag_tensor_sigma: bool = False):
    """Calculate the term P4 for paramagnetic sublattice.
    For details see documentation "Integrated intensity from powder diffraction".
    """
    sigma_11 = tensor_sigma[0]
    sigma_12 = tensor_sigma[1]
    sigma_21 = tensor_sigma[3]
    sigma_22 = tensor_sigma[4]
    
    p_4 = sigma_21.real*sigma_11.imag - sigma_21.imag*sigma_11.real + \
          sigma_22.real*sigma_12.imag - sigma_22.imag*sigma_12.real

    dder = {}
    if flag_tensor_sigma:
        ones = numpy.ones_like(sigma_11.real)
        dder_sigma_11_real = - sigma_21.imag * ones
        dder_sigma_11_imag = sigma_21.real * ones
        dder_sigma_12_real = - sigma_22.imag * ones
        dder_sigma_12_imag = sigma_22.real * ones
        dder_sigma_21_real = sigma_11.imag * ones
        dder_sigma_21_imag = - sigma_11.real * ones
        dder_sigma_22_real = sigma_12.imag * ones
        dder_sigma_22_imag = - sigma_12.real * ones
        zeros = numpy.zeros_like(dder_sigma_11_real)
        dder["tensor_sigma_real"] = numpy.stack([
            dder_sigma_11_real, dder_sigma_12_real, zeros, 
            dder_sigma_21_real, dder_sigma_22_real, zeros, zeros, zeros, zeros], axis=0)
        dder["tensor_sigma_imag"] = numpy.stack([
            dder_sigma_11_imag, dder_sigma_12_imag, zeros, 
            dder_sigma_21_imag, dder_sigma_22_imag, zeros, zeros, zeros, zeros], axis=0)
    return p_4, dder


def calc_chiral_term_cos_cube_para(tensor_sigma, flag_tensor_sigma: bool = False):
    """Calculate the term P5 for paramagnetic sublattice.
    For details see documentation "Integrated intensity from powder diffraction".
    """
    sigma_13 = tensor_sigma[2]
    sigma_23 = tensor_sigma[5]
    
    p_5 = 2*(sigma_23.real*sigma_13.imag - sigma_23.imag*sigma_13.real)

    dder = {}
    if flag_tensor_sigma:
        ones = numpy.ones_like(sigma_13.real)
        dder_sigma_13_real = -2 * sigma_23.imag * ones
        dder_sigma_13_imag = 2 * sigma_23.real * ones
        dder_sigma_23_real = 2 * sigma_13.imag * ones
        dder_sigma_23_imag = -2 * sigma_13.real * ones
        zeros = numpy.zeros_like(dder_sigma_13_real)
        dder["tensor_sigma_real"] = numpy.stack([
            zeros, zeros, dder_sigma_13_real, 
            zeros, zeros, dder_sigma_23_real, zeros, zeros, zeros], axis=0)
        dder["tensor_sigma_imag"] = numpy.stack([
            zeros, zeros, dder_sigma_13_imag, 
            zeros, zeros, dder_sigma_23_imag, zeros, zeros, zeros], axis=0)
    return p_5, dder


def calc_cross_term_ordered(f_nucl, f_m_perp, flag_f_nucl: bool = False, flag_f_m_perp: bool = False):
    """Calculate the term O1 for the magnetically ordered sublattice.
    For details see documentation "Integrated intensity from powder diffraction".
    """
    f_m_perp_z = f_m_perp[2]
    
    o_1 = 2*(f_nucl.real*f_m_perp_z.real + f_nucl.imag*f_m_perp_z.imag)

    dder = {}
    if flag_f_nucl:
        dder["f_nucl_real"] = 2 * f_m_perp_z.real * numpy.ones_like(f_nucl.real)
        dder["f_nucl_imag"] = 2 * f_m_perp_z.imag * numpy.ones_like(f_nucl.imag)
    if flag_f_m_perp:
        dder_f_m_perp_z_real = 2 * f_nucl.real * numpy.ones_like(f_m_perp_z.real)
        dder_f_m_perp_z_imag = 2 * f_nucl.imag * numpy.ones_like(f_m_perp_z.imag)
        zeros = numpy.zeros_like(dder_f_m_perp_z_real)
        dder["f_m_perp_real"] = numpy.stack([zeros, zeros, dder_f_m_perp_z_real], axis=0)
        dder["f_m_perp_imag"] = numpy.stack([zeros, zeros, dder_f_m_perp_z_imag], axis=0)
    return o_1, dder


def calc_chiral_term_ordered(f_m_perp, flag_f_m_perp: bool = False):
    """Calculate the term O2 for the magnetically ordered sublattice.
    For details see documentation "Integrated intensity from powder diffraction".
    """
    f_m_perp_x = f_m_perp[0]
    f_m_perp_y = f_m_perp[1]
    
    o_2 = 2*(f_m_perp_y.real*f_m_perp_x.imag - f_m_perp_y.imag*f_m_perp_x.real)

    dder = {}
    if flag_f_m_perp:
        dder_f_m_perp_x_real = -2 * f_m_perp_y.imag * numpy.ones_like(f_m_perp_x.real)
        dder_f_m_perp_x_imag = 2 * f_m_perp_y.real * numpy.ones_like(f_m_perp_x.imag)
        dder_f_m_perp_y_real = 2 * f_m_perp_x.imag * numpy.ones_like(f_m_perp_y.real)
        dder_f_m_perp_y_imag = -2 * f_m_perp_x.real * numpy.ones_like(f_m_perp_y.imag)
        zeros = numpy.zeros_like(dder_f_m_perp_x_real)
        dder["f_m_perp_real"] = numpy.stack([dder_f_m_perp_x_real, dder_f_m_perp_y_real, zeros], axis=0)
        dder["f_m_perp_imag"] = numpy.stack([dder_f_m_perp_x_imag, dder_f_m_perp_y_imag, zeros], axis=0)
    return o_2, dder


def calc_m_sq_mix(tensor_sigma, f_m_perp, flag_tensor_sigma: bool = False, flag_f_m_perp: bool = False):
    """Calculate the term M1 for the case of coexistiong paramatic and magnetically ordered sublattices.
    For details see documentation "Integrated intensity from powder diffraction".

    tensor_sigma describe paramagnetic sublattice
    f_m_perp describe ordered sublattice
    """
    sigma_13 = tensor_sigma[2]
    sigma_23 = tensor_sigma[5]
    f_m_perp_x = f_m_perp[0]
    f_m_perp_y = f_m_perp[1]
    
    m_1 = 2*(sigma_13.real*f_m_perp_x.real + sigma_13.imag*f_m_perp_x.imag + 
            sigma_23.real*f_m_perp_y.real + sigma_23.imag*f_m_perp_y.imag)

    dder = {}
    if flag_tensor_sigma:
        dder_sigma_13_real = 2*f_m_perp_x.real*numpy.ones_like(sigma_13.real)
        dder_sigma_13_imag = 2*f_m_perp_x.imag*numpy.ones_like(sigma_13.imag)
        dder_sigma_23_real = 2*f_m_perp_y.real*numpy.ones_like(sigma_23.real)
        dder_sigma_23_imag = 2*f_m_perp_y.imag*numpy.ones_like(sigma_23.imag)
        zeros = numpy.zeros_like(dder_sigma_13_real)
        dder["tensor_sigma_real"] = numpy.stack([
            zeros, zeros, dder_sigma_13_real,
            zeros, zeros, dder_sigma_23_real,
            zeros, zeros, zeros], axis=0)
        dder["tensor_sigma_imag"] = numpy.stack([
            zeros, zeros, dder_sigma_13_imag,
            zeros, zeros, dder_sigma_23_imag,
            zeros, zeros, zeros], axis=0)
    if flag_f_m_perp:
        dder_f_m_perp_x_real = 2 * sigma_13.real * numpy.ones_like(f_m_perp_x.real)
        dder_f_m_perp_x_imag = 2 * sigma_13.imag * numpy.ones_like(f_m_perp_x.imag)
        dder_f_m_perp_y_real = 2 * sigma_23.real * numpy.ones_like(f_m_perp_y.real)
        dder_f_m_perp_y_imag = 2 * sigma_13.imag * numpy.ones_like(f_m_perp_y.imag)
        zeros = numpy.zeros_like(dder_f_m_perp_x_real)
        dder["f_m_perp_real"] = numpy.stack([dder_f_m_perp_x_real, dder_f_m_perp_y_real, zeros], axis=0)
        dder["f_m_perp_imag"] = numpy.stack([dder_f_m_perp_x_imag, dder_f_m_perp_y_imag, zeros], axis=0)
    return m_1, dder


def calc_chiral_term_sin_sq_mix(
        tensor_sigma, f_m_perp, flag_tensor_sigma: bool = False, flag_f_m_perp: bool = False):
    """Calculate the term M2 for the case of coexistiong paramatic and magnetically ordered sublattices.
    For details see documentation "Integrated intensity from powder diffraction".

    tensor_sigma describe paramagnetic sublattice
    f_m_perp describe ordered sublattice
    """
    sigma_12 = tensor_sigma[1]
    sigma_21 = tensor_sigma[3]
    f_m_perp_z = f_m_perp[2]
    
    m_2 = sigma_12.real*f_m_perp_z.imag - sigma_12.imag*f_m_perp_z.real + \
        sigma_21.imag*f_m_perp_z.real - sigma_21.real*f_m_perp_z.imag

    dder = {}
    if flag_tensor_sigma:
        dder_sigma_12_real =  f_m_perp_z.imag*numpy.ones_like(sigma_12.real)
        dder_sigma_12_imag = -f_m_perp_z.real*numpy.ones_like(sigma_12.imag)
        dder_sigma_21_real = -f_m_perp_z.imag*numpy.ones_like(sigma_21.real)
        dder_sigma_21_imag =  f_m_perp_z.real*numpy.ones_like(sigma_21.imag)
        zeros = numpy.zeros_like(dder_sigma_12_real)
        dder["tensor_sigma_real"] = numpy.stack([
            zeros, dder_sigma_12_real, zeros,
            dder_sigma_21_real, zeros, zeros,
            zeros, zeros, zeros], axis=0)
        dder["tensor_sigma_imag"] = numpy.stack([
            zeros, dder_sigma_12_imag, zeros,
            dder_sigma_21_imag, zeros, zeros,
            zeros, zeros, zeros], axis=0)
    if flag_f_m_perp:
        dder_f_m_perp_z_real = (-sigma_12.imag + sigma_21.imag) * numpy.ones_like(f_m_perp_z.real)
        dder_f_m_perp_z_imag = ( sigma_12.real + sigma_21.real) * numpy.ones_like(f_m_perp_z.imag)
        zeros = numpy.zeros_like(dder_f_m_perp_z_real)
        dder["f_m_perp_real"] = numpy.stack([zeros, zeros, dder_f_m_perp_z_real], axis=0)
        dder["f_m_perp_imag"] = numpy.stack([zeros, zeros, dder_f_m_perp_z_imag], axis=0)
    return m_2, dder


def calc_chiral_term_cos_sq_mix(
        tensor_sigma, f_m_perp, flag_tensor_sigma: bool = False, flag_f_m_perp: bool = False):
    """Calculate the term M3 for the case of coexistiong paramatic and magnetically ordered sublattices.
    For details see documentation "Integrated intensity from powder diffraction".

    tensor_sigma describe paramagnetic sublattice
    f_m_perp describe ordered sublattice
    """
    sigma_13 = tensor_sigma[2]
    sigma_23 = tensor_sigma[5]
    f_m_perp_x = f_m_perp[0]
    f_m_perp_y = f_m_perp[1]
    
    m_3 = 2.*(
        sigma_23.real*f_m_perp_x.imag - sigma_23.imag*f_m_perp_x.real + \
        sigma_13.imag*f_m_perp_y.real - sigma_13.real*f_m_perp_y.imag)

    dder = {}
    if flag_tensor_sigma:
        dder_sigma_13_real = -f_m_perp_y.imag*numpy.ones_like(sigma_13.real)
        dder_sigma_13_imag =  f_m_perp_y.real*numpy.ones_like(sigma_13.imag)
        dder_sigma_23_real =  f_m_perp_x.imag*numpy.ones_like(sigma_23.real)
        dder_sigma_23_imag = -f_m_perp_x.real*numpy.ones_like(sigma_23.imag)
        zeros = numpy.zeros_like(dder_sigma_13_real)
        dder["tensor_sigma_real"] = numpy.stack([
            zeros, zeros, dder_sigma_13_real,
            zeros, zeros, dder_sigma_23_real, 
            zeros, zeros, zeros], axis=0)
        dder["tensor_sigma_imag"] = numpy.stack([
            zeros, zeros, dder_sigma_13_imag,
            zeros, zeros, dder_sigma_23_imag,
            zeros, zeros, zeros], axis=0)
    if flag_f_m_perp:
        dder_f_m_perp_x_real = -2 * sigma_23.imag * numpy.ones_like(f_m_perp_x.real)
        dder_f_m_perp_x_imag =  2 * sigma_23.real * numpy.ones_like(f_m_perp_x.imag)
        dder_f_m_perp_y_real =  2 * sigma_13.imag * numpy.ones_like(f_m_perp_y.real)
        dder_f_m_perp_y_imag = -2 * sigma_13.real * numpy.ones_like(f_m_perp_y.imag)
        zeros = numpy.zeros_like(dder_f_m_perp_x_real)
        dder["f_m_perp_real"] = numpy.stack([dder_f_m_perp_x_real, dder_f_m_perp_y_real, zeros], axis=0)
        dder["f_m_perp_imag"] = numpy.stack([dder_f_m_perp_x_imag, dder_f_m_perp_y_imag, zeros], axis=0)
    return m_3, dder


def calc_powder_iint_1d_para(f_nucl, tensor_sigma, polarization, flipper, magnetic_field,
        flag_f_nucl: bool = False, flag_tensor_sigma: bool = False,
        flag_polarization: bool = False, flag_flipper: bool = False):
    """Calculated powderly averaged integrated intensity for paramagnetic sublattice in equatorial plane (alpha=90 deg.)
    For details see documentation "Integrated intensity from powder diffraction".

    Output is two integrated intensities (plus and minus) with derivatives.
    """
    p_u = polarization
    p_d = polarization*(2.*flipper-1.)    

    f_nucl_sq = numpy.square(numpy.abs(f_nucl))
    p_1, dder_p_1 = calc_m_sq_sin_sq_para(tensor_sigma, flag_tensor_sigma=flag_tensor_sigma)
    p_3, dder_p_3 = calc_cross_term_para(f_nucl, tensor_sigma, flag_f_nucl=flag_f_nucl,
        flag_tensor_sigma=flag_tensor_sigma)
    magnetic_field_sq = numpy.square(magnetic_field)
    iint_plus = f_nucl_sq + magnetic_field_sq * p_1 + p_u * magnetic_field*p_3
    iint_minus = f_nucl_sq + magnetic_field_sq * p_1 - p_d * magnetic_field*p_3

    dder_plus = {}
    dder_minus = {}
    if flag_polarization:
        dder_plus["polarization"] = magnetic_field*p_3*numpy.ones_like(polarization)
        dder_minus["polarization"] = -1*magnetic_field*p_3*(2*flipper-1)*numpy.ones_like(polarization)
    if flag_flipper:
        dder_minus["flipper"] = -2*magnetic_field*p_3*polarization*numpy.ones_like(flipper)
    if flag_f_nucl:
        dder_plus["f_nucl_real"] = 2 * f_nucl.real + p_u * magnetic_field * dder_p_3["f_nucl_real"]
        dder_plus["f_nucl_imag"] = 2 * f_nucl.imag + p_u * magnetic_field * dder_p_3["f_nucl_imag"]
        dder_minus["f_nucl_real"] = 2 * f_nucl.real - p_d * magnetic_field * dder_p_3["f_nucl_real"]
        dder_minus["f_nucl_imag"] = 2 * f_nucl.imag - p_d * magnetic_field * dder_p_3["f_nucl_imag"]
    if flag_tensor_sigma:
        dder_plus["tensor_sigma_real"] = magnetic_field_sq * dder_p_1["tensor_sigma_real"] + p_u * magnetic_field*dder_p_3["tensor_sigma_real"]
        dder_plus["tensor_sigma_imag"] = magnetic_field_sq * dder_p_1["tensor_sigma_imag"] + p_u * magnetic_field*dder_p_3["tensor_sigma_imag"]
        dder_minus["tensor_sigma_real"] = magnetic_field_sq * dder_p_1["tensor_sigma_real"] - p_d * magnetic_field*dder_p_3["tensor_sigma_real"]
        dder_minus["tensor_sigma_imag"] = magnetic_field_sq * dder_p_1["tensor_sigma_imag"] - p_d * magnetic_field*dder_p_3["tensor_sigma_imag"]
    return iint_plus, iint_minus, dder_plus, dder_minus


def calc_powder_iint_1d_ordered(f_nucl, f_m_perp,
        flag_f_nucl: bool = False, flag_f_m_perp: bool = False):
    """Calculated powderly averaged integrated intensity for ordered sublattice in equatorial plane (alpha=90 deg.)
    For details see documentation "Integrated intensity from powder diffraction".

    In equatorial plane the scattering is not a function of polarized neutrons.
    Output is integrated intensities with derivatives.
    """
    f_nucl_sq = numpy.square(numpy.abs(f_nucl))
    f_m_perp_sq = numpy.square(numpy.abs(f_m_perp)).sum(axis=0)
    iint = f_nucl_sq + f_m_perp_sq

    dder = {}
    if flag_f_nucl:
        dder["f_nucl_real"] = 2 * f_nucl.real
        dder["f_nucl_imag"] = 2 * f_nucl.imag
    if flag_f_m_perp:
        dder["f_m_perp_real"] = 2*f_m_perp.real
        dder["f_m_perp_imag"] = 2*f_m_perp.imag
    return iint, dder


def calc_powder_iint_1d_mix(f_nucl, f_m_perp, tensor_sigma, polarization, flipper, magnetic_field,
        flag_f_nucl: bool = False, flag_f_m_perp_ordered: bool = False, flag_tensor_sigma: bool = False,
        flag_polarization: bool = False, flag_flipper: bool = False,
        flag_magnetic_field: bool = False):
    """Calculated powderly averaged integrated intensity in case of coexisting paramagnetic and magnetically ordered sublattice at
    scattering in equatorial plane (alpha=90 deg.).
    For details see documentation "Integrated intensity from powder diffraction".

    Output is two integrated intensities (plus and minus) with derivatives.
    """
    p_u = polarization
    p_d = polarization*(2.*flipper-1.)
    iint_plus_para, iint_minus_para, dder_plus_para, dder_minus_para = calc_powder_iint_1d_para(
        f_nucl, tensor_sigma, polarization, flipper, magnetic_field,
        flag_f_nucl=flag_f_nucl, flag_tensor_sigma=flag_tensor_sigma,
        flag_polarization=flag_polarization, flag_flipper=flag_flipper)

    f_m_perp_sq = numpy.square(numpy.abs(f_m_perp)).sum(axis=0)

    m_2, dder_m_2 = calc_chiral_term_sin_sq_mix(
        tensor_sigma, f_m_perp, flag_tensor_sigma=flag_tensor_sigma, flag_f_m_perp=flag_f_m_perp_ordered)

    iint_plus = iint_plus_para + f_m_perp_sq + p_u * magnetic_field * m_2
    iint_minus = iint_minus_para + f_m_perp_sq + p_d * magnetic_field * m_2

    dder_plus = {}
    dder_minus = {}
    if flag_f_nucl:
        dder_plus["f_nucl_real"] = dder_plus_para["f_nucl_real"] 
        dder_minus["f_nucl_real"] = dder_minus_para["f_nucl_real"] 
    if flag_f_m_perp_ordered:
        dder_plus["f_m_perp_ordered_real"] = 2*f_m_perp.real + p_u * magnetic_field * dder_m_2["f_m_perp_real"]
        dder_plus["f_m_perp_ordered_imag"] = 2*f_m_perp.imag + p_u * magnetic_field * dder_m_2["f_m_perp_imag"]
        dder_minus["f_m_perp_ordered_real"] = 2*f_m_perp.real + p_d * magnetic_field * dder_m_2["f_m_perp_real"]
        dder_minus["f_m_perp_ordered_imag"] = 2*f_m_perp.imag + p_d * magnetic_field * dder_m_2["f_m_perp_imag"]
    if flag_tensor_sigma:
        dder_plus["tensor_sigma_real"] = dder_plus_para["tensor_sigma_real"] + p_u * magnetic_field * dder_m_2["tensor_sigma_real"]
        dder_plus["tensor_sigma_imag"] = dder_plus_para["tensor_sigma_imag"] + p_u * magnetic_field * dder_m_2["tensor_sigma_imag"]
        dder_minus["tensor_sigma_real"] = dder_minus_para["tensor_sigma_real"] + p_d * magnetic_field * dder_m_2["tensor_sigma_real"]
        dder_minus["tensor_sigma_imag"] = dder_minus_para["tensor_sigma_imag"] + p_d * magnetic_field * dder_m_2["tensor_sigma_imag"]
    if flag_polarization:
        dder_plus["polarization"] = dder_plus_para["polarization"] + magnetic_field * m_2 * numpy.ones_like(polarization)
        dder_minus["polarization"] = dder_minus_para["polarization"] + magnetic_field * m_2 * (2.*flipper-1.)* numpy.ones_like(polarization)
    if flag_flipper:
        dder_minus["flipper"] = dder_minus_para["flippe"] + magnetic_field * m_2 * 2*polarization* numpy.ones_like(flipper)
    if flag_magnetic_field:
        dder_plus["magnetic_field"] = dder_plus_para["magnetic_field"] + p_u* m_2 * numpy.ones_like(magnetic_field)
        dder_minus["magnetic_field"] = dder_minus_para["magnetic_field"] + p_d* m_2 * numpy.ones_like(magnetic_field)
    return iint_plus, iint_minus, dder_plus, dder_minus


def calc_powder_iint_2d_para(f_nucl, tensor_sigma, beam_polarization, flipper_efficiency, magnetic_field,
        alpha_det, dict_in_out,
        flag_f_nucl: bool = False, flag_tensor_sigma: bool = False,
        flag_polarization: bool = False, flag_flipper: bool = False):
    """Calculated powderly averaged integrated intensity for paramagnetic sublattice in equatorial plane (alpha=90 deg.)
    For details see documentation "Integrated intensity from powder diffraction".

    Output is two integrated intensities (plus and minus) with derivatives.
    """
    p_u = beam_polarization
    p_d = beam_polarization*(2.*flipper_efficiency-1.)    

    s_a_sq = numpy.square(numpy.sin(alpha_det))[:, :, na]
    c_a = numpy.cos(alpha_det)[:, :, na]

    f_nucl_sq = numpy.square(numpy.abs(f_nucl))
    p_1, dder_p_1 = calc_m_sq_sin_sq_para(tensor_sigma, flag_tensor_sigma=flag_tensor_sigma)
    p_2, dder_p_2 = calc_m_sq_cos_sq_para(tensor_sigma, flag_tensor_sigma=flag_tensor_sigma)
    p_3, dder_p_3 = calc_cross_term_para(f_nucl, tensor_sigma, flag_f_nucl=flag_f_nucl,
        flag_tensor_sigma=flag_tensor_sigma)
    p_4, dder_p_4 = calc_chiral_term_cos_sin_sq_para(tensor_sigma, flag_tensor_sigma=flag_tensor_sigma)
    p_5, dder_p_5 = calc_chiral_term_cos_cube_para(tensor_sigma, flag_tensor_sigma=flag_tensor_sigma)
    magnetic_field_sq = numpy.square(magnetic_field)
    dict_in_out["p_1"] = p_1
    dict_in_out["p_2"] = p_2
    dict_in_out["p_3"] = p_3
    dict_in_out["p_4"] = p_4
    dict_in_out["p_5"] = p_5

    iint_plus = (f_nucl_sq + magnetic_field_sq * p_2)[na, na, :] + \
        (magnetic_field_sq * p_1 - magnetic_field_sq * p_2 + p_u * magnetic_field*p_3)[na, na, :] * s_a_sq + \
         (p_u * magnetic_field_sq)[na, na, :] * ((p_4-p_5)*s_a_sq + p_5[na, na, :])*c_a
    iint_minus = (f_nucl_sq + magnetic_field_sq * p_2)[na, na, :] + \
        (magnetic_field_sq * p_1 - magnetic_field_sq * p_2 - p_d * magnetic_field*p_3)[na, na, :] * s_a_sq - \
        (p_d * magnetic_field_sq)[na, na, :] * ((p_4-p_5)[na, na, :]*s_a_sq + p_5[na, na, :])*c_a

    dder_plus = {}
    dder_minus = {}
    if flag_polarization:
        dder_plus["beam_polarization"] = magnetic_field*p_3*s_a_sq + magnetic_field_sq * ((p_4-p_5)*s_a_sq + p_5)*c_a 
        dder_minus["beam_polarization"] = -(2*flipper_efficiency-1)*dder_plus["beam_polarization"]
    if flag_flipper:
        dder_minus["flipper_efficiency"] = -2*beam_polarization*(magnetic_field*p_3*s_a_sq + magnetic_field_sq * ((p_4-p_5)*s_a_sq + p_5)*c_a)
    if flag_f_nucl:
        dder_plus["f_nucl_real"] = 2 * f_nucl.real + p_u * magnetic_field * s_a_sq * dder_p_3["f_nucl_real"]
        dder_plus["f_nucl_imag"] = 2 * f_nucl.imag + p_u * magnetic_field * s_a_sq * dder_p_3["f_nucl_imag"]
        dder_minus["f_nucl_real"] = 2 * f_nucl.real - p_d * magnetic_field * s_a_sq * dder_p_3["f_nucl_real"]
        dder_minus["f_nucl_imag"] = 2 * f_nucl.imag - p_d * magnetic_field * s_a_sq * dder_p_3["f_nucl_imag"]
    if flag_tensor_sigma:
        dder_plus["tensor_sigma_real"] = None
        dder_plus["tensor_sigma_imag"] = None
        dder_minus["tensor_sigma_real"] = None
        dder_minus["tensor_sigma_imag"] = None
    return iint_plus, iint_minus, dder_plus, dder_minus


def calc_powder_iint_2d_ordered(f_nucl, f_m_perp, beam_polarization, flipper_efficiency,
        alpha_det, dict_in_out, 
        flag_f_nucl: bool = False, flag_f_m_perp: bool = False,
        flag_polarization: bool = False, flag_flipper: bool = False):
    """Calculated powderly averaged integrated intensity for paramagnetic sublattice in equatorial plane (alpha=90 deg.)
    For details see documentation "Integrated intensity from powder diffraction".

    Output is two integrated intensities (plus and minus) with derivatives.
    """
    p_u = beam_polarization
    p_d = beam_polarization*(2.*flipper_efficiency-1.)    

    s_a_sq = numpy.square(numpy.sin(alpha_det))[:, :, na]
    c_a = numpy.cos(alpha_det)[:, :, na]

    f_nucl_sq = numpy.square(numpy.abs(f_nucl))
    f_m_perp_sq = ((f_m_perp *numpy.conjugate(f_m_perp)).real).sum(axis=0)
    o_1, dder_o_1 = calc_cross_term_ordered(f_nucl, f_m_perp, flag_f_nucl=flag_f_nucl, flag_f_m_perp=flag_f_m_perp)
    o_2, dder_o_2 = calc_chiral_term_ordered(f_m_perp, flag_f_m_perp=flag_f_m_perp)
    dict_in_out["o_1"] = o_1
    dict_in_out["o_2"] = o_2

    iint_plus = (f_nucl_sq + f_m_perp_sq)[na, na, :] + p_u * (o_1+o_2)[na, na, :] * c_a
    iint_minus = (f_nucl_sq + f_m_perp_sq)[na, na, :] - p_d * (o_1+o_2)[na, na, :] * c_a

    dder_plus = {}
    dder_minus = {}
    if flag_polarization:
        dder_plus["beam_polarization"] = (o_1 + o_2) * c_a
        dder_minus["beam_polarization"] = -(2*flipper_efficiency - 1) * dder_plus["beam_polarization"]
    if flag_flipper:
        dder_minus["flipper_efficiency"] = -beam_polarization * (o_1 + o_2) * c_a
    if flag_f_nucl:
        dder_plus["f_nucl_real"] = 2 * f_nucl.real 
        dder_plus["f_nucl_imag"] = 2 * f_nucl.imag 
        dder_minus["f_nucl_real"] = 2 * f_nucl.real 
        dder_minus["f_nucl_imag"] = 2 * f_nucl.imag 
    if flag_f_m_perp:
        dder_plus["tensor_sigma_real"] = None
        dder_plus["tensor_sigma_imag"] = None
        dder_minus["tensor_sigma_real"] = None
        dder_minus["tensor_sigma_imag"] = None
    return iint_plus, iint_minus, dder_plus, dder_minus



def calc_powder_iint_2d_mix(f_nucl, tensor_sigma, f_m_perp_ordered, beam_polarization, flipper_efficiency, magnetic_field,
        alpha_det, dict_in_out_phase,
        flag_f_nucl: bool = False, flag_tensor_sigma: bool = False, flag_f_m_perp_ordered: bool = False,
        flag_polarization: bool = False, flag_flipper: bool = False):
    """Calculated powderly averaged integrated intensity for paramagnetic sublattice in equatorial plane (alpha=90 deg.)
    For details see documentation "Integrated intensity from powder diffraction".

    Output is two integrated intensities (plus and minus) with derivatives.
    """
    p_u = beam_polarization
    p_d = beam_polarization*(2.*flipper_efficiency-1.)    

    s_a_sq = numpy.square(numpy.sin(alpha_det))[:, :, na]
    c_a = numpy.cos(alpha_det)[:, :, na]

    f_nucl_sq = numpy.square(numpy.abs(f_nucl))
    p_1, dder_p_1 = calc_m_sq_sin_sq_para(tensor_sigma, flag_tensor_sigma=flag_tensor_sigma)
    p_2, dder_p_2 = calc_m_sq_cos_sq_para(tensor_sigma, flag_tensor_sigma=flag_tensor_sigma)
    p_3, dder_p_3 = calc_cross_term_para(f_nucl, tensor_sigma, flag_f_nucl=flag_f_nucl,
        flag_tensor_sigma=flag_tensor_sigma)
    p_4, dder_p_4 = calc_chiral_term_cos_sin_sq_para(tensor_sigma, flag_tensor_sigma=flag_tensor_sigma)
    p_5, dder_p_5 = calc_chiral_term_cos_cube_para(tensor_sigma, flag_tensor_sigma=flag_tensor_sigma)
    magnetic_field_sq = numpy.square(magnetic_field)
    dict_in_out_phase["p_1"] = p_1
    dict_in_out_phase["p_2"] = p_2
    dict_in_out_phase["p_3"] = p_3
    dict_in_out_phase["p_4"] = p_4
    dict_in_out_phase["p_5"] = p_5

    f_m_perp_sq = ((f_m_perp_ordered *numpy.conjugate(f_m_perp_ordered)).real).sum(axis=0)
    o_1, dder_o_1 = calc_cross_term_ordered(f_nucl, f_m_perp_ordered, flag_f_nucl=flag_f_nucl, flag_f_m_perp=flag_f_m_perp_ordered)
    o_2, dder_o_2 = calc_chiral_term_ordered(f_m_perp_ordered, flag_f_m_perp=flag_f_m_perp_ordered)
    dict_in_out_phase["o_1"] = o_1
    dict_in_out_phase["o_2"] = o_2

    m_1, dder_m_1 = calc_m_sq_mix(
        tensor_sigma, f_m_perp_ordered, flag_tensor_sigma=flag_tensor_sigma, flag_f_m_perp=flag_f_m_perp_ordered)
    m_2, dder_m_2 = calc_chiral_term_sin_sq_mix(
        tensor_sigma, f_m_perp_ordered, flag_tensor_sigma=flag_tensor_sigma, flag_f_m_perp=flag_f_m_perp_ordered)
    m_3, dder_m_3 = calc_chiral_term_cos_sq_mix(
        tensor_sigma, f_m_perp_ordered, flag_tensor_sigma=flag_tensor_sigma, flag_f_m_perp=flag_f_m_perp_ordered)

    dict_in_out_phase["m_1"] = m_1
    dict_in_out_phase["m_2"] = m_2
    dict_in_out_phase["m_3"] = m_3

    iint_plus = (f_nucl_sq + magnetic_field_sq * p_2 + f_m_perp_sq)[na, na, :] + p_u*magnetic_field*m_3+\
        (magnetic_field_sq * p_1 - magnetic_field_sq * p_2 + p_u * magnetic_field*p_3 + p_u*magnetic_field*m_2-p_u*magnetic_field*m_3)[na, na, :] * s_a_sq + \
         (p_u * magnetic_field_sq)[na, na, :] * ((p_4-p_5)*s_a_sq + p_5[na, na, :])*c_a + \
         + (p_u*o_1+p_u*o_2+magnetic_field*m_1)[na, na, :] * c_a 
    iint_minus = (f_nucl_sq + magnetic_field_sq * p_2 + f_m_perp_sq)[na, na, :] - p_d*magnetic_field*m_3+\
        (magnetic_field_sq * p_1 - magnetic_field_sq * p_2 - p_d * magnetic_field*p_3 - p_d*magnetic_field*m_2-p_d*magnetic_field*m_3)[na, na, :] * s_a_sq - \
        (p_d * magnetic_field_sq)[na, na, :] * ((p_4-p_5)[na, na, :]*s_a_sq + p_5[na, na, :])*c_a +\
        (-p_d*o_1-p_d*o_2+magnetic_field*m_1)[na, na, :] * c_a

    dder_plus = {}
    dder_minus = {}
    return iint_plus, iint_minus, dder_plus, dder_minus



"""
Expressions to work with symmetry elements

symm_elems: numpy.shape[13, n_elems]
[numerator_x, numerator_y, numerator_z, denominator_xyz,
 r_11, r_12, r_13,
 r_21, r_22, r_23,
 r_31, r_32, r_33]

mag_symm_elems: [22, n_elems]
[numerator_x, numerator_y, numerator_z, denominator_xyz,
 r_11, r_12, r_13,
 r_21, r_22, r_23,
 r_31, r_32, r_33,
 m_11, m_12, m_13,
 m_21, m_22, m_23,
 m_31, m_32, m_33]

For details see documentation.
"""
import numpy
from .function_1_strings import transform_string_to_digits
from cryspy.A_functions_base.function_2_sym_elems import calc_numerators_and_common_denominator

from cryspy.A_functions_base.matrix_operations import calc_m1_m2, calc_m_v

na = numpy.newaxis


def calc_symm_flags(symm_elems, atom_symm_elems):
    # cond_1: (R_s - E) R_a = 0
    bn_s_1, bn_s_2, bn_s_3, bd_s = symm_elems[0], symm_elems[1], symm_elems[2], symm_elems[3]
    rse_11 = symm_elems[4] - 1
    rse_12 = symm_elems[5]
    rse_13 = symm_elems[6]
    rse_21 = symm_elems[7]
    rse_22 = symm_elems[8] - 1
    rse_23 = symm_elems[9]
    rse_31 = symm_elems[10]
    rse_32 = symm_elems[11]
    rse_33 = symm_elems[12] - 1

    bn_a_1, bn_a_2, bn_a_3, bd_a = atom_symm_elems[0], atom_symm_elems[1], \
        atom_symm_elems[2], atom_symm_elems[3]
    if (atom_symm_elems.shape[0]==4):
        cond_1_11 = numpy.zeros_like(rse_33*bn_a_1) == 0
        cond_1_12 = cond_1_11
        cond_1_13 = cond_1_11
        cond_1_21 = cond_1_11
        cond_1_22 = cond_1_11
        cond_1_23 = cond_1_11
        cond_1_31 = cond_1_11
        cond_1_32 = cond_1_11
        cond_1_33 = cond_1_11
    else:
        ra_11, ra_12, ra_13, ra_21, ra_22, ra_23, ra_31, ra_32, ra_33 = \
            atom_symm_elems[4], atom_symm_elems[5], atom_symm_elems[6], \
            atom_symm_elems[7], atom_symm_elems[8], atom_symm_elems[9], \
            atom_symm_elems[10], atom_symm_elems[11], atom_symm_elems[12]

        cond_1_11 = rse_11*ra_11 + rse_12*ra_21 + rse_13*ra_31 == 0
        cond_1_12 = rse_11*ra_12 + rse_12*ra_22 + rse_13*ra_32 == 0
        cond_1_13 = rse_11*ra_13 + rse_12*ra_23 + rse_13*ra_33 == 0
        cond_1_21 = rse_21*ra_11 + rse_22*ra_21 + rse_23*ra_31 == 0
        cond_1_22 = rse_21*ra_12 + rse_22*ra_22 + rse_23*ra_32 == 0
        cond_1_23 = rse_21*ra_13 + rse_22*ra_23 + rse_23*ra_33 == 0
        cond_1_31 = rse_31*ra_11 + rse_32*ra_21 + rse_33*ra_31 == 0
        cond_1_32 = rse_31*ra_12 + rse_32*ra_22 + rse_33*ra_32 == 0
        cond_1_33 = rse_31*ra_13 + rse_32*ra_23 + rse_33*ra_33 == 0

    # cond_2: (R_s - E) b_a + b_s = int
    
    bd_sa = bd_s*bd_a

    p1_1 = bd_s * (rse_11*bn_a_1 + rse_12*bn_a_2 + rse_13*bn_a_3) + bd_a * bn_s_1
    p1_2 = bd_s * (rse_21*bn_a_1 + rse_22*bn_a_2 + rse_23*bn_a_3) + bd_a * bn_s_2
    p1_3 = bd_s * (rse_31*bn_a_1 + rse_32*bn_a_2 + rse_33*bn_a_3) + bd_a * bn_s_3
    
    cond_2_1 = numpy.mod(p1_1, bd_sa) == 0
    cond_2_2 = numpy.mod(p1_2, bd_sa) == 0
    cond_2_3 = numpy.mod(p1_3, bd_sa) == 0
    
    la = numpy.logical_and
    symm_flag_point_to_point = \
        la(cond_1_11, la(cond_1_12, la(cond_1_13, la(
           cond_1_21, la(cond_1_22, la(cond_1_23, la(
           cond_1_31, la(cond_1_32, la(cond_1_33, la(
           cond_2_1, la(cond_2_2, cond_2_3)))))))))))
    return symm_flag_point_to_point


def calc_multiplicity_by_atom_symm_elems(full_symm_elems, atom_symm_elems):
    """Calculate the multiplicity for atoms defined by symm_elems and atom_symm_atoms.
    
    atom_symm_atoms can be composed as only from atom position and also from atom position and symmetry elements
    """
    symm_flag_point_to_point = calc_symm_flags(
        full_symm_elems[:, :, na], atom_symm_elems[:, na, :])
    point_multiplicity = numpy.floor_divide(
        full_symm_elems.shape[1], numpy.sum(symm_flag_point_to_point, axis=0))
    return point_multiplicity


def calc_asymmetric_unit_cell_indexes(n_abc, full_symm_elems):
    """
    Calculate indexes of asymmetric unit cell.

    Input parameters:
        - symmetry elements;
        - points number.

    Multiplication of points number on corresponding symmetry element should
    give integer number.

    """
    n_a, n_b, n_c = n_abc[0], n_abc[1], n_abc[2]

    point_index = numpy.stack(numpy.meshgrid(
        numpy.arange(n_a), numpy.arange(n_b), numpy.arange(n_c),
        indexing="ij"), axis=0)
    point_index = point_index.reshape(point_index.shape[0], numpy.prod(point_index.shape[1:]))
    
    elem_r = full_symm_elems[4:13]
    elem_b = full_symm_elems[:4]

    r_ind = calc_m_v(
        numpy.expand_dims(elem_r, axis=1),
        numpy.expand_dims(point_index, axis=2), flag_m=False, flag_v=False)[0]

    div, mod = numpy.divmod(numpy.expand_dims(n_abc, axis=1), numpy.expand_dims(elem_b[3], axis=0))
    if not(numpy.all(mod == 0)):
        raise KeyError("Symmetry elements do not match with number of points")
    point_index_s = numpy.mod(r_ind + numpy.expand_dims(div * elem_b[:3], axis=1),
        numpy.expand_dims(numpy.expand_dims(n_abc, axis=1), axis=2))
    value_index_s = n_c*n_b*point_index_s[0] + n_c*point_index_s[1] + point_index_s[2]
    value_index_s_sorted = numpy.sort(value_index_s, axis=1)

    a, ind_a_u_c, counts_a_u_c = numpy.unique(
        value_index_s_sorted[:, 0], return_index=True, return_counts=True)

    point_index_s_a_u_c = point_index[:, ind_a_u_c]

    return point_index_s_a_u_c, counts_a_u_c


def sum_elem_symm_b(elem_symm_b1, elem_symm_b2):
    num_1, denom_1 = elem_symm_b1[:3], elem_symm_b1[3:4]
    num_2, denom_2 = elem_symm_b2[:3], elem_symm_b2[3:4]
    num = num_1 * denom_2 + num_2 * denom_1
    denom = denom_1*denom_2
    gcd = (numpy.gcd(num, denom)).min(axis=0)
    denom_out = numpy.floor_divide(denom, gcd)
    
    num_out = numpy.floor_divide(num, gcd)
    res = numpy.concatenate([num_out, denom_out], axis=0)
    return res


def calc_full_mag_elems(mag_elems_o, mag_elems_c):
    r_o = mag_elems_o[4:13,:]
    r_c = mag_elems_c[4:13,:]
    b_n_o = mag_elems_o[:3,:]
    b_d_o = mag_elems_o[3:4,:]
    b_c = mag_elems_c[:4,:]
    theta_o = mag_elems_o[13:14,:]
    theta_c = mag_elems_c[13:14,:]
    
    
    theta_fs = theta_o[:, :, na] * theta_c[:, na, :]
    r_fs, dder_r_fs = calc_m1_m2(
        r_o[:, :, na], r_c[:, na, :],
        flag_m1=False, flag_m2=False)
    
    rb, dder_rb = calc_m_v(r_c[:, na, :], b_n_o[:, :, na], flag_m=False, flag_v=False)
    b_d_o = numpy.broadcast_arrays(b_c[0, na, :], b_d_o[:, :, na])[1]
    
    se_rb = numpy.concatenate([rb, b_d_o], axis=0)

    b_fs = sum_elem_symm_b(se_rb, b_c[:, na, :])
    res = numpy.concatenate([b_fs, r_fs, theta_fs], axis=0)
    res_2d = res.reshape((res.shape[0], numpy.prod(res.shape[1:])), order="C")
    mag_elems_fs = res_2d # May be it is not enough and unique elements should be given
    # mag_elems_fs = numpy.unique(res_2d, axis=1)
    return mag_elems_fs


def take_symm_elem_by_string(string):
    """
    Define symm_elems from string

    string: x,z+1/2,-y+3/4
    symm_elems: 0 2 3 4 1 0 0 0 0 1 0 -1 0
    """
    labels=("x", "y", "z")
    l_name = "".join(string.strip().split()).lstrip("(").rstrip(")").split(",")
    rij, bi_nd = [], []
    for _name in l_name:
        coefficients, offset = transform_string_to_digits(_name, labels)
        rij.extend([coefficients[0].numerator, coefficients[1].numerator, coefficients[2].numerator])
        bi_nd.extend([offset.numerator, offset.denominator])
    b_nd = calc_numerators_and_common_denominator(*bi_nd)
    symm_elem = numpy.array(b_nd+tuple(rij), dtype=int)
    return symm_elem


def take_symm_elems_by_string(string):
    l_symm_elems = []
    for s_val in string:
        symm_elem = take_symm_elem_by_string(s_val)
        l_symm_elems.append(symm_elem)
    symm_elems = numpy.stack(l_symm_elems).transpose()
    return symm_elems


def apply_symm_elems_to_index_xyz(symm_elems, index_xyz, points_abc):
    """Apply symmetry elements to indexes."""
    b_n_1, b_n_2, b_n_3 = symm_elems[0], symm_elems[1], symm_elems[2]
    b_d = symm_elems[3]
    r_11, r_12, r_13 = symm_elems[4], symm_elems[5], symm_elems[6]
    r_21, r_22, r_23 = symm_elems[7], symm_elems[8], symm_elems[9]
    r_31, r_32, r_33 = symm_elems[10], symm_elems[11], symm_elems[12]
    i_1, i_2, i_3 = index_xyz[0], index_xyz[1], index_xyz[2]
    
    n1, n2, n3 = points_abc[0], points_abc[1], points_abc[2]
    p_1, p_2, p_3 = n1//b_d, n2//b_d, n3//b_d
    
    ni_1 = numpy.mod(r_11*i_1 + r_12*i_2 + r_13*i_3 + b_n_1*p_1, n1)
    ni_2 = numpy.mod(r_21*i_1 + r_22*i_2 + r_23*i_3 + b_n_2*p_2, n2)
    ni_3 = numpy.mod(r_31*i_1 + r_32*i_2 + r_33*i_3 + b_n_3*p_3, n3)
    ni = numpy.stack([ni_1, ni_2, ni_3], axis=0)
    return ni


def sum_b_elems(b_1, b_2):
    b_1_x, b_1_y, b_1_z, b_1_d = b_1[0], b_1[1], b_1[2], b_1[3]
    b_2_x, b_2_y, b_2_z, b_2_d = b_2[0], b_2[1], b_2[2], b_2[3]
    denom_common = numpy.lcm.reduce([b_1_d, b_2_d])
    c_1 = denom_common // b_1_d
    c_2 = denom_common // b_2_d 
    b_s_x = (b_1_x*c_1 + b_2_x*c_2)%denom_common
    b_s_y = (b_1_y*c_1 + b_2_y*c_2)%denom_common
    b_s_z = (b_1_z*c_1 + b_2_z*c_2)%denom_common
    b_s = numpy.stack([b_s_x, b_s_y, b_s_z, denom_common])
    b_sum = b_s // numpy.expand_dims(numpy.gcd.reduce(b_s), axis=0)
    return b_sum


def calc_full_symm_elems_by_reduced(reduced_symm_elems, centrosymmetry, centrosymmetry_position, translation_elems):
    l_fse = []    
    for te in translation_elems.transpose():
        fse_1 = numpy.copy(reduced_symm_elems)
        te_exp = numpy.ones_like(fse_1[:4,:])*numpy.expand_dims(te, axis=1)
        fse_1[:4,:] = sum_b_elems(fse_1[:4,:], te_exp)
        l_fse.append(fse_1)
    fse_a = numpy.concatenate(l_fse, axis=1)

    if centrosymmetry:
        dcp = numpy.copy(centrosymmetry_position)
        dcp[:3] = 2 * dcp[:3]
        fse_c = numpy.copy(fse_a)
        fse_c[:3] = -1*fse_c[:3]
        fse_c[4:] = -1*fse_c[4:]
        dcp_exp = numpy.ones_like(fse_c[:4,:]) * numpy.expand_dims(dcp, axis=1)
        fse_c[:4,:] = sum_b_elems(dcp_exp, fse_c[:4,:])
        full_symm_elems = numpy.concatenate([fse_a, fse_c], axis=1)
    else:
        full_symm_elems = fse_a
    return full_symm_elems


def calc_equivalent_reflections(index_hkl, reduced_symm_elems, centrosymmetry: bool = False):
    r_11, r_12, r_13 = numpy.expand_dims(reduced_symm_elems[4], axis=0), numpy.expand_dims(reduced_symm_elems[5], axis=0), numpy.expand_dims(reduced_symm_elems[6], axis=0)
    r_21, r_22, r_23 = numpy.expand_dims(reduced_symm_elems[7], axis=0), numpy.expand_dims(reduced_symm_elems[8], axis=0), numpy.expand_dims(reduced_symm_elems[9], axis=0)
    r_31, r_32, r_33 = numpy.expand_dims(reduced_symm_elems[10], axis=0), numpy.expand_dims(reduced_symm_elems[11], axis=0), numpy.expand_dims(reduced_symm_elems[12], axis=0)

    h, k, l = numpy.expand_dims(index_hkl[0], axis=1), numpy.expand_dims(index_hkl[1], axis=1), numpy.expand_dims(index_hkl[2], axis=1)

    index_hkl_equivalent = numpy.stack([
        r_11*h + r_21*k + r_31*l,
        r_12*h + r_22*k + r_32*l,
        r_13*h + r_23*k + r_33*l], axis=0)

    if centrosymmetry:
        index_hkl_equivalent = numpy.concatenate([index_hkl_equivalent, -1*index_hkl_equivalent], axis=2)
    return index_hkl_equivalent


def define_centring_type_by_symm_elems(symm_elems):
    """Define centring_type by symmetry elements."""
    
    symm_elem_I = numpy.array([1,1,1,2,1,0,0,0,1,0,0,0,1], dtype=int)
    symm_elem_A = numpy.array([0,1,1,2,1,0,0,0,1,0,0,0,1], dtype=int)
    symm_elem_B = numpy.array([1,0,1,2,1,0,0,0,1,0,0,0,1], dtype=int)
    symm_elem_C = numpy.array([1,1,0,2,1,0,0,0,1,0,0,0,1], dtype=int)
    symm_elem_H_1 = numpy.array([2,1,0,3,1,0,0,0,1,0,0,0,1], dtype=int)
    symm_elem_H_2 = numpy.array([1,2,0,3,1,0,0,0,1,0,0,0,1], dtype=int)
    symm_elem_R_1 = numpy.array([2,1,1,3,1,0,0,0,1,0,0,0,1], dtype=int)
    symm_elem_R_2 = numpy.array([1,2,2,3,1,0,0,0,1,0,0,0,1], dtype=int)
    symm_elem_Rrev_1 = numpy.array([1,2,1,3,1,0,0,0,1,0,0,0,1], dtype=int)
    symm_elem_Rrev_2 = numpy.array([2,1,2,3,1,0,0,0,1,0,0,0,1], dtype=int)
    
    flag_I = numpy.any(numpy.all(symm_elems[:13, :] == numpy.expand_dims(symm_elem_I, axis=1), axis=0))
    flag_A = numpy.any(numpy.all(symm_elems[:13, :] == numpy.expand_dims(symm_elem_A, axis=1), axis=0))
    flag_B = numpy.any(numpy.all(symm_elems[:13, :] == numpy.expand_dims(symm_elem_B, axis=1), axis=0))
    flag_C = numpy.any(numpy.all(symm_elems[:13, :] == numpy.expand_dims(symm_elem_C, axis=1), axis=0))
    flag_H_1 = numpy.any(numpy.all(symm_elems[:13, :] == numpy.expand_dims(symm_elem_H_1, axis=1), axis=0))
    flag_H_2 = numpy.any(numpy.all(symm_elems[:13, :] == numpy.expand_dims(symm_elem_H_2, axis=1), axis=0))
    flag_R_1 = numpy.any(numpy.all(symm_elems[:13, :] == numpy.expand_dims(symm_elem_R_1, axis=1), axis=0))
    flag_R_2 = numpy.any(numpy.all(symm_elems[:13, :] == numpy.expand_dims(symm_elem_R_2, axis=1), axis=0))
    flag_Rrev_1 = numpy.any(numpy.all(symm_elems[:13, :] == numpy.expand_dims(symm_elem_Rrev_1, axis=1), axis=0))
    flag_Rrev_2 = numpy.any(numpy.all(symm_elems[:13, :] == numpy.expand_dims(symm_elem_Rrev_2, axis=1), axis=0))
    
    flag_F = numpy.all([flag_A, flag_B, flag_C])
    flag_H = numpy.all([flag_H_1, flag_H_2])
    flag_R = numpy.all([flag_R_1, flag_R_2])
    flag_Rrev = numpy.all([flag_Rrev_1, flag_Rrev_2])
    
    if flag_F:
        centring_type = "F"
    elif flag_H:
        centring_type = "H"
    elif flag_R:
        centring_type = "R"
    elif flag_Rrev:
        centring_type = "Rrev"
    elif flag_I:
        centring_type = "I"
    elif flag_A:
        centring_type = "A"
    elif flag_B:
        centring_type = "B"
    elif flag_C:
        centring_type = "C"
    else:
        centring_type = "P"
    return centring_type


def define_bravais_type_by_symm_elems(symm_elems):
    n_symm_elems = symm_elems.shape[1]
    
    symm_elem_inversion = numpy.array([-1,0,0,0,-1,0,0,0,-1], dtype=int)
    
    centrosymmetry = numpy.any(numpy.all(symm_elems[4:13, :] == numpy.expand_dims(symm_elem_inversion, axis=1), axis=0))
    n_centrosymmetry = 2 if centrosymmetry else 1
    
    centring_type = define_centring_type_by_symm_elems(symm_elems)
    
    d_centring_type = {"P":1, "A":2, "B":2, "C":2, "F":4, "H":3, "I":2, "R":3, "Rrev":3}
    n_centring_type = d_centring_type[centring_type]
    
    n_elems = n_symm_elems // (n_centrosymmetry * n_centring_type)
    
    if n_symm_elems % (n_centrosymmetry * n_centring_type) != 0:
        raise UserWarning("Error in number of symmetry elements")
        
    
    flag_1y = numpy.any(symm_elems[5,:] != 0)
    flag_1z = numpy.any(symm_elems[6,:] != 0)
    flag_2x = numpy.any(symm_elems[7,:] != 0)
    flag_2z = numpy.any(symm_elems[9,:] != 0)
    flag_3x = numpy.any(symm_elems[10,:] != 0)
    flag_3y = numpy.any(symm_elems[11,:] != 0)

    flag_1y_no = numpy.all(symm_elems[5,:] == 0)
    flag_1z_no = numpy.all(symm_elems[6,:] == 0)
    flag_2x_no = numpy.all(symm_elems[7,:] == 0)
    flag_2z_no = numpy.all(symm_elems[9,:] == 0)
    flag_3x_no = numpy.all(symm_elems[10,:] == 0)
    flag_3y_no = numpy.all(symm_elems[11,:] == 0)

    
    flag_cubic = numpy.all([flag_1y, flag_1z, flag_2x, flag_2z, flag_3x, flag_3y])
    flag_no_mix = numpy.all([flag_1y_no, flag_1z_no, flag_2x_no, flag_2z_no, flag_3x_no, flag_3y_no])
    flag_rhombohedral = numpy.all([flag_cubic, centring_type.startswith("P"), n_elems == 6])
    if flag_rhombohedral:
        flag_cubic = False
    
    flag_triclinic = n_elems == 1
    flag_monoclinic = n_elems == 2
    flag_orthorombic = n_elems == 4
    flag_tetragonal = n_elems == 8
    
    flag_hexagonal_trigonal = numpy.all([not(flag_rhombohedral), n_elems%3==0])
    
    it_coordinate_system_code = None

    if flag_cubic:
        bravais_type = "c"+centring_type[0]
    elif flag_rhombohedral:
        bravais_type = "h"+centring_type[0]
        it_coordinate_system_code = "r"# FIXME: not sure
    elif flag_tetragonal:
        bravais_type = "t"+centring_type[0]
    elif flag_orthorombic:
        if centring_type in ["P", "I", "F"]:
            bravais_type = "o"+centring_type[0]
        else:
            bravais_type = "oS"
    elif flag_monoclinic:
        if centring_type in ["P", "I", "F"]:
            bravais_type = "m"+centring_type[0]
        else:
            bravais_type = "mS"
    elif flag_triclinic:
        bravais_type = "a"+centring_type[0]
    elif flag_hexagonal_trigonal:
        bravais_type = "h"+centring_type[0]
        it_coordinate_system_code = "h"
    else:
        bravais_type = None
    
    return bravais_type, it_coordinate_system_code


"""Description of Phase and PhaseL classes."""
from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class Phase(ItemN):
    """Phase class.

    Attributes
    ----------
        - label (mandatory)
        - scale, igsize, u, v, w, x, y (optional)
    """

    ATTR_MANDATORY_NAMES = ("label",)
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("label", )

    ATTR_OPTIONAL_NAMES = ("scale", "igsize", "x", "y", "u", "v", "w")
    ATTR_OPTIONAL_TYPES = (float, float, float, float, float, float, float)
    ATTR_OPTIONAL_CIF = ("scale", "igsize", "X", "Y", "U", "V", "W")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("scale", "igsize", "x", "y", "u", "v", "w")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}

    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "phase"

    def __init__(self, **kwargs) -> NoReturn:
        super(Phase, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"igsize": 0., "x": 0., "y": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def get_dictionary(self):
        res = {}
        res["phase_name"] = numpy.array([self.label,], dtype=str)
        n_items = 1
        if self.is_attribute("u"):
            p_u = numpy.array([self.u,], dtype=float)
            r_u = numpy.array([self.u_refinement,], dtype=bool)
        else:
            p_u = numpy.zeros((n_items,), dtype=float)
            r_u = numpy.zeros((n_items,), dtype=bool)

        if self.is_attribute("v"):
            p_v = numpy.array([self.v,], dtype=float)
            r_v = numpy.array([self.v_refinement,], dtype=bool)
        else:
            p_v = numpy.zeros((n_items,), dtype=float)
            r_v = numpy.zeros((n_items,), dtype=bool)

        if self.is_attribute("w"):
            p_w = numpy.array([self.w,], dtype=float)
            r_w = numpy.array([self.w_refinement,], dtype=bool)
        else:
            p_w = numpy.zeros((n_items,), dtype=float)
            r_w = numpy.zeros((n_items,), dtype=bool)

        if self.is_attribute("x"):
            p_x = numpy.array([self.x,], dtype=float)
            r_x = numpy.array([self.x_refinement,], dtype=bool)
        else:
            p_x = numpy.zeros((n_items,), dtype=float)
            r_x = numpy.zeros((n_items,), dtype=bool)

        if self.is_attribute("y"):
            p_y = numpy.array([self.y,], dtype=float)
            r_y = numpy.array([self.y_refinement,], dtype=bool)
        else:
            p_y = numpy.zeros((n_items,), dtype=float)
            r_y = numpy.zeros((n_items,), dtype=bool)

        res["phase_resolution_parameters"] = numpy.stack([p_u, p_v, p_w, p_x, p_y], axis=0)
        res["flags_phase_resolution_parameters"] = numpy.stack([r_u, r_v, r_w, r_x, r_y], axis=0)

        if self.is_attribute("igsize"):
            res["phase_ig"] = numpy.array([self.igsize,], dtype=float)
            res["flags_phase_ig"] = numpy.array([self.igsize_refinement,], dtype=bool)
        else:
            res["phase_ig"] = numpy.zeros((n_items,), dtype=float)
            res["flags_phase_ig"] = numpy.zeros((n_items,), dtype=bool)

        if self.is_attribute("scale"):
            res["phase_scale"] = numpy.array([self.scale,], dtype=float)
            res["flags_phase_scale"] = numpy.array([self.scale_refinement,], dtype=bool)
        else:
            res["phase_scale"] = numpy.zeros((n_items,), dtype=float)
            res["flags_phase_scale"] = numpy.zeros((n_items,), dtype=bool)
        return res


class PhaseL(LoopN):
    """Phases of the sample.

    Attributes
    ----------
        - label (mandatory)
        - scale, igsize, u, v, w, x, y (optional)
    """

    ITEM_CLASS = Phase
    ATTR_INDEX = "label"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(PhaseL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def get_dictionary(self):
        res = {}
        res["phase_name"] = numpy.array(self.label, dtype=str)
        n_items = len(self.items)
        if self.is_attribute("u"):
            p_u = numpy.array(self.u, dtype=float)
            r_u = numpy.array(self.u_refinement, dtype=bool)
        else:
            p_u = numpy.zeros((n_items,), dtype=float)
            r_u = numpy.zeros((n_items,), dtype=bool)

        if self.is_attribute("v"):
            p_v = numpy.array(self.v, dtype=float)
            r_v = numpy.array(self.v_refinement, dtype=bool)
        else:
            p_v = numpy.zeros((n_items,), dtype=float)
            r_v = numpy.zeros((n_items,), dtype=bool)

        if self.is_attribute("w"):
            p_w = numpy.array(self.w, dtype=float)
            r_w = numpy.array(self.w_refinement, dtype=bool)
        else:
            p_w = numpy.zeros((n_items,), dtype=float)
            r_w = numpy.zeros((n_items,), dtype=bool)

        if self.is_attribute("x"):
            p_x = numpy.array(self.x, dtype=float)
            r_x = numpy.array(self.x_refinement, dtype=bool)
        else:
            p_x = numpy.zeros((n_items,), dtype=float)
            r_x = numpy.zeros((n_items,), dtype=bool)

        if self.is_attribute("y"):
            p_y = numpy.array(self.y, dtype=float)
            r_y = numpy.array(self.y_refinement, dtype=bool)
        else:
            p_y = numpy.zeros((n_items,), dtype=float)
            r_y = numpy.zeros((n_items,), dtype=bool)

        res["phase_resolution_parameters"] = numpy.stack([p_u, p_v, p_w, p_x, p_y], axis=0)
        res["flags_phase_resolution_parameters"] = numpy.stack([r_u, r_v, r_w, r_x, r_y], axis=0)

        if self.is_attribute("igsize"):
            res["phase_ig"] = numpy.array(self.igsize, dtype=float)
            res["flags_phase_ig"] = numpy.array(self.igsize_refinement, dtype=bool)
        else:
            res["phase_ig"] = numpy.zeros((n_items,), dtype=float)
            res["flags_phase_ig"] = numpy.zeros((n_items,), dtype=bool)

        if self.is_attribute("scale"):
            res["phase_scale"] = numpy.array(self.scale, dtype=float)
            res["flags_phase_scale"] = numpy.array(self.scale_refinement, dtype=bool)
        else:
            res["phase_scale"] = numpy.zeros((n_items,), dtype=float)
            res["flags_phase_scale"] = numpy.zeros((n_items,), dtype=bool)
        return res



import numpy
import scipy
import scipy.special


def calc_rho_normalized(radius, coeff, zeta, n, kappa=1):
    """
    Give normalized Slatter type function.

        R = ((2*zeta0)**(n0+0.5)/((2*n0)!)**0.5)*r**(n0-1)*exp(-zeta0*r)
        (see Clementi_1974)
    """
    zeta_ang = kappa*zeta/0.52918
    hh = numpy.power(2.*zeta_ang, n.astype(float)+0.5)
    coeff_norm = hh / numpy.sqrt(scipy.special.factorial(2*n))
    r_nm1 = numpy.power(
        numpy.expand_dims(radius, axis=1),
        numpy.expand_dims(n-1, axis=0))
    norm_density = (numpy.expand_dims(coeff*coeff_norm, axis=0)*r_nm1*
                    numpy.exp(-numpy.expand_dims(zeta_ang, axis=0)*
                              numpy.expand_dims(radius, axis=1))).sum(axis=1)
    return norm_density


def calc_density_spherical(radius, shell_population, shell_coeff, shell_zeta, shell_n, kappa):
    density_spherical = numpy.zeros_like(radius)
    for population, coeff, zeta, n in zip(shell_population, shell_coeff, shell_zeta, shell_n):
        rho_normalized = calc_rho_normalized(radius, coeff, zeta, n, kappa=kappa)
        density_spherical += 4.*numpy.pi*population*numpy.square(rho_normalized)
    return density_spherical


def calc_transs(l_max, nn, zeta, sthovl):
    """ integral( r**nn * exp(-zeta r) * j_l(Qr) dr) / Q**l
    Q = 4 pi sint / lambda
    # ATTENTION: in one of my note it was written factor r**2 but it looks that it was mistake
    """
    Q = 4*numpy.pi*sthovl
    a = [0. for h in range(17)]
    ff = [0 for h in range(l_max+1)]
    d = Q**2 + zeta**2
    a[0] = 0.0
    a[1] = 1./d
    n = nn-1
    tz = 2.*zeta
    ts = 4.*numpy.pi
    for l in range(l_max+1):
        ll = l+1
        if (ll != 1):
            a[ll] = a[ll-1]*ts*l*1./d
            a[ll-1] = numpy.zeros_like(sthovl) 
        for nx in range(ll, n+1):
            i1 = nx
            i2 = nx+1
            i3 = i2+1
            a[i3-1] = (tz*nx*a[i2-1] - (nx+l)*(nx-ll)*a[i1-1])*1./d
        ff[ll-1] = a[i3-1]
    return ff


def calc_jl_per_2sthovll(sthovl, coeff, n, zeta, kappa=1, l_max = 3):
    c_ij = numpy.expand_dims(coeff, axis=1) * numpy.expand_dims(coeff, axis=0)
    zeta_ang = kappa*zeta/0.529177
    hh = numpy.power(2.*zeta_ang, n.astype(float)+0.5)
    coeff_norm = hh / numpy.sqrt(scipy.special.factorial(2*n))
    coeff_norm_ij = numpy.expand_dims(coeff_norm, axis=1) * numpy.expand_dims(coeff_norm, axis=0)
    
    nn_ij = numpy.expand_dims(n, axis=1) + numpy.expand_dims(n, axis=0) # FIXME: why not -2?
    zeta_ij = numpy.expand_dims(zeta_ang, axis=1) + numpy.expand_dims(zeta_ang, axis=0)

    res = numpy.zeros(sthovl.shape+(l_max+1,))

    for c, cc, nn, zz in zip(c_ij.flatten(), coeff_norm_ij.flatten(), nn_ij.flatten(), zeta_ij.flatten()):
        ff = calc_transs(l_max, nn, zz, sthovl)
        ff_l = numpy.stack(ff, axis=1)
        
        val = c*cc*ff_l
        res += val
    return res


def calc_jl(sthovl, coeff, n, zeta, kappa=1, l_max = 3):
    # q_l = numpy.power(numpy.expand_dims(4*numpy.pi*sthovl, axis=1), numpy.expand_dims(numpy.arange(l_max+1), axis=0))
    q_l = numpy.power(numpy.expand_dims(2.*sthovl, axis=1), numpy.expand_dims(numpy.arange(l_max+1), axis=0))
    jl_per_2sthovll = calc_jl_per_2sthovll(sthovl, coeff, n, zeta, kappa=kappa, l_max=l_max)
    jl = jl_per_2sthovll*q_l
    return jl


"""
Base crystallographic functions.

Functions
---------
    - calc_beta_by_u
    - vibration_constraints,
    - apply_constraint_on_cell_by_type_cell

They are stored in the list FUNCTIONS
"""
from typing import Tuple
import numpy


def calc_beta_by_u(u_i, cell):
    r"""
    Calculate beta.

    u_11, u_22, u_33, u_12, u_13, u_23 = u_i

    cell is Cell class

    Output:
        $\beta_{11}$, $\beta_{22}$, $\beta_{33}$
        $\beta_{12}$, $\beta_{13}$, $\beta_{23}$
    """
    (u_11, u_22, u_33, u_12, u_13, u_23) = u_i

    ia = cell.reciprocal_length_a
    ib = cell.reciprocal_length_b
    ic = cell.reciprocal_length_c
    beta_11 = 2.*numpy.pi**2*u_11*ia**2
    beta_22 = 2.*numpy.pi**2*u_22*ib**2
    beta_33 = 2.*numpy.pi**2*u_33*ic**2
    beta_12 = 2.*numpy.pi**2*u_12*ia*ib
    beta_13 = 2.*numpy.pi**2*u_13*ia*ic
    beta_23 = 2.*numpy.pi**2*u_23*ib*ic
    return beta_11, beta_22, beta_33, beta_12, beta_13, beta_23


def vibration_constraints(numb, param_i, sigma_i, ref_i):
    """
    Constraints on atomic vibrations.

    Parameters
    ----------
    numb : TYPE
        DESCRIPTION.
    param_i : TYPE
        DESCRIPTION.
    sigma_i : TYPE
        DESCRIPTION.
    ref_i : TYPE
        DESCRIPTION.

    Returns
    -------
    param_i : TYPE
        DESCRIPTION.
    sigma_i : TYPE
        DESCRIPTION.
    ref_i : TYPE
        DESCRIPTION.
    constr_i : TYPE
        DESCRIPTION.

    """
    p_11, p_22, p_33, p_12, p_13, p_23 = param_i
    s_11, s_22, s_33, s_12, s_13, s_23 = sigma_i
    p_11_ref, p_22_ref, p_33_ref, p_12_ref, p_13_ref, p_23_ref = ref_i
    p_11_constr, p_22_constr, p_33_constr = False, False, False
    p_12_constr, p_13_constr, p_23_constr = False, False, False

    if numb == 1:
        p_12 = 0.
        s_12 = 0.
        p_12_ref = False
        p_12_constr = True
        p_23 = 0.
        s_23 = 0.
        p_23_ref = False
        p_23_constr = True
    elif numb == 2:
        p_13 = 0.
        s_13 = 0.
        p_13_ref = False
        p_13_constr = True
        p_23 = 0.
        s_23 = 0.
        p_23_ref = False
        p_23_constr = True
    elif numb == 3:
        p_12 = 0.
        s_12 = 0.
        p_12_ref = False
        p_12_constr = True
        p_13 = 0.
        s_13 = 0.
        p_13_ref = False
        p_13_constr = True
    elif numb == 4:
        p_12 = 0.
        s_12 = 0.
        p_12_ref = False
        p_12_constr = True
        p_13 = 0.
        s_13 = 0.
        p_13_ref = False
        p_13_constr = True
        p_23 = 0.
        s_23 = 0.
        p_23_ref = False
        p_23_constr = True
    elif numb == 5:
        p_22 = p_11
        s_22 = s_11
        p_22_ref = False
        p_22_constr = True
        p_13 = 0.
        s_13 = 0.
        p_13_ref = False
        p_13_constr = True
        p_23 = 0.
        s_23 = 0.
        p_23_ref = False
        p_23_constr = True
    elif numb == 6:
        p_22 = p_11
        s_22 = s_11
        p_22_ref = False
        p_22_constr = True
        p_23 = p_13
        s_23 = s_13
        p_23_ref = False
        p_23_constr = True
    elif numb == 7:
        p_22 = p_11
        s_22 = s_11
        p_22_ref = False
        p_22_constr = True
        p_23 = -1.*p_13
        s_23 = s_13
        p_23_ref = False
        p_23_constr = True
    elif numb == 8:
        p_22 = p_11
        s_22 = s_11
        p_22_ref = False
        p_22_constr = True
        p_12 = 0.
        s_12 = 0.
        p_12_ref = False
        p_12_constr = True
        p_13 = 0.
        s_13 = 0.
        p_13_ref = False
        p_13_constr = True
        p_23 = 0.
        s_23 = 0.
        p_23_ref = False
        p_23_constr = True
    elif numb == 9:
        p_33 = p_22
        s_33 = s_22
        p_33_ref = False
        p_33_constr = True
        p_12 = 0.
        s_12 = 0.
        p_12_ref = False
        p_12_constr = True
        p_13 = 0.
        s_13 = 0.
        p_13_ref = False
        p_13_constr = True
    elif numb == 10:
        p_33 = p_22
        s_33 = s_22
        p_33_ref = False
        p_33_constr = True
        p_13 = p_12
        s_13 = s_12
        p_13_ref = False
        p_13_constr = True
    elif numb == 11:
        p_33 = p_22
        s_33 = s_22
        p_33_ref = False
        p_33_constr = True
        p_13 = -1.*p_12
        s_13 = s_12
        p_13_ref = False
        p_13_constr = True
    elif numb == 12:
        p_33 = p_22
        s_33 = s_22
        p_33_ref = False
        p_33_constr = True
        p_12 = 0.
        s_12 = 0.
        p_12_ref = False
        p_12_constr = True
        p_13 = 0.
        s_13 = 0.
        p_13_ref = False
        p_13_constr = True
        p_23 = 0.
        s_23 = 0.
        p_23_ref = False
        p_23_constr = True
    elif numb == 13:
        p_12 = 0.5*p_22         
        s_12 = 0.5*s_22         
        p_12_ref = False
        p_12_constr = True
        p_23 = 0.
        s_23 = 0.
        p_23_ref = False
        p_23_constr = True
    elif numb == 14:
        p_12 = 0.5*p_22
        s_12 = 0.5*s_22
        p_12_ref = False
        p_12_constr = True
        p_13 = 0.
        s_13 = 0.
        p_13_ref = False
        p_13_constr = True
        p_23 = 0.
        s_23 = 0.
        p_23_ref = False
        p_23_constr = True
    elif numb == 15:
        p_12 = 0.5*p_22
        s_12 = 0.5*s_22
        p_12_ref = False
        p_12_constr = True
        p_23 = 2.0*p_13
        s_23 = 2.0*s_13
        p_23_ref = False
        p_23_constr = True
    elif numb == 16:
        p_22 = p_11
        s_22 = s_11
        p_22_ref = False
        p_22_constr = True
        p_12 = 0.5*p_11      
        s_12 = 0.5*s_11      
        p_12_ref = False
        p_12_constr = True
        p_13 = 0.
        s_13 = 0.
        p_13_ref = False
        p_13_constr = True
        p_23 = 0.
        s_23 = 0.
        p_23_ref = False
        p_23_constr = True
    elif numb == 17:
        p_22 = p_11
        s_22 = s_11
        p_22_ref = False
        p_22_constr = True
        p_33 = p_11
        s_33 = s_11
        p_33_ref = False
        p_33_constr = True
        p_12 = 0.
        s_12 = 0.
        p_12_ref = False
        p_12_constr = True
        p_13 = 0.
        s_13 = 0.
        p_13_ref = False
        p_13_constr = True
        p_23 = 0.
        s_23 = 0.
        p_23_ref = False
        p_23_constr = True
    elif numb == 18:
        p_22 = p_11
        s_22 = s_11
        p_22_ref = False
        p_22_constr = True
        p_33 = p_11
        s_33 = s_11
        p_33_ref = False
        p_33_constr = True
        p_13 = p_12
        s_13 = s_12
        p_13_ref = False
        p_13_constr = True
        p_23 = p_12
        s_23 = s_12
        p_23_ref = False
        p_23_constr = True

    param_i = (p_11, p_22, p_33, p_12, p_13, p_23)
    sigma_i = (s_11, s_22, s_33, s_12, s_13, s_23)
    ref_i = (p_11_ref, p_22_ref, p_33_ref, p_12_ref, p_13_ref, p_23_ref)
    constr_i = (p_11_constr, p_22_constr, p_33_constr, p_12_constr,
                p_13_constr, p_23_constr)
    return param_i, sigma_i, ref_i, constr_i


def apply_constraint_on_cell_by_type_cell(
        cell_param: Tuple[float], cell_sigma: Tuple[float],
        cell_ref: Tuple[bool], type_cell: str,
        it_coordinate_system_code: str) -> (Tuple[float], Tuple[float],
                                            Tuple[bool], Tuple[bool]):
    """
       Apply constraints.

    Parameters
    ----------
    cell_param : Tuple[float]
        (a, b, c, alpha, beta, gamma).
    cell_sigma : Tuple[float]
        (a_sigma, b_sigma, c_sigma, alpha_sigma, beta_sigma, gamma_sigma).
    cell_ref : Tuple[bool]
        (a_sigma, b_sigma, c_sigma, alpha_sigma, beta_sigma, gamma_sigma).
    type_cell : str
        DESCRIPTION.
    it_coordinate_system_code : str
        DESCRIPTION.

    Returns
    -------
    (Tuple[float], Tuple[float], Tuple[bool], Tuple[bool])
        cell_param, cell_sigma, cell_ref, cell_constr.

    """
    (a, b, c, alpha, beta, gamma) = cell_param
    (a_sig, b_sig, c_sig, alpha_sig, beta_sig, gamma_sig) = cell_sigma
    (a_ref, b_ref, c_ref, alpha_ref, beta_ref, gamma_ref) = cell_ref
    a_constr, b_constr, c_constr, alpha_constr, beta_constr, gamma_constr = \
        False, False, False, False, False, False

    if type_cell == "aP":
        pass
    elif type_cell.startswith("m"):
        alpha, alpha_sig, alpha_ref, alpha_constr = 90., 0., False, True
        gamma, gamma_sig, gamma_ref, gamma_constr = 90., 0., False, True
    elif type_cell.startswith("o"):
        alpha, alpha_sig, alpha_ref, alpha_constr = 90., 0., False, True
        beta, beta_sig, beta_ref, beta_constr = 90., 0., False, True
        gamma, gamma_sig, gamma_ref, gamma_constr = 90., 0., False, True
    elif ((type_cell.startswith("t"))):  # FIXME: check  | (type_cell == "hP")
        b, b_sig, b_ref, b_constr = a, a_sig, False, True
        alpha, alpha_sig, alpha_ref, alpha_constr = 90., 0., False, True
        beta, beta_sig, beta_ref, beta_constr = 90., 0., False, True
        gamma, gamma_sig, gamma_ref, gamma_constr = 90., 0., False, True
    elif ((type_cell.startswith("hP"))):
        if it_coordinate_system_code.lower() == "h":
            b, b_sig, b_ref, b_constr = a, a_sig, False, True
            alpha, alpha_sig, alpha_ref, alpha_constr = 90., 0., False, True
            beta, beta_sig, beta_ref, beta_constr = 90., 0., False, True
            gamma, gamma_sig, gamma_ref, gamma_constr = 120., 0., False, True
        else:
            b, b_sig, b_ref, b_constr = a, a_sig, False, True
            c, c_sig, c_ref, c_constr = a, a_sig, False, True
            beta, beta_sig, beta_ref, beta_constr = alpha, alpha_sig, \
                False, True
            gamma, gamma_sig, gamma_ref, gamma_constr = alpha, alpha_sig,\
                False, True
    elif (type_cell == "hR"):
        if it_coordinate_system_code.lower() == "h":
            b, b_sig, b_ref, b_constr = a, a_sig, False, True
            alpha, alpha_sig, alpha_ref, alpha_constr = 90., 0., False, True
            beta, beta_sig, beta_ref, beta_constr = 90., 0., False, True
            gamma, gamma_sig, gamma_ref, gamma_constr = 120., 0., False, True
        else:
            b, b_sig, b_ref, b_constr = a, a_sig, False, True
            c, c_sig, c_ref, c_constr = a, a_sig, False, True
            beta, beta_sig, beta_ref, beta_constr = alpha, alpha_sig, \
                False, True
            gamma, gamma_sig, gamma_ref, gamma_constr = alpha, alpha_sig,\
                False, True
    elif type_cell.startswith("c"):
        b, b_sig, b_ref, b_constr = a, a_sig, False, True
        c, c_sig, c_ref, c_constr = a, a_sig, False, True
        alpha, alpha_sig, alpha_ref, alpha_constr = 90., 0., False, True
        beta, beta_sig, beta_ref, beta_constr = 90., 0., False, True
        gamma, gamma_sig, gamma_ref, gamma_constr = 90., 0., False, True
    cell_param = (a, b, c, alpha, beta, gamma)
    cell_sigma = (a_sig, b_sig, c_sig, alpha_sig, beta_sig, gamma_sig)
    cell_ref = (a_ref, b_ref, c_ref, alpha_ref, beta_ref, gamma_ref)
    cell_constr = (a_constr, b_constr, c_constr, alpha_constr, beta_constr,
                   gamma_constr)
    return cell_param, cell_sigma, cell_ref, cell_constr



import os
import numpy
import scipy
import scipy.optimize

from cryspy.A_functions_base.symmetry_elements import \
    calc_asymmetric_unit_cell_indexes

from cryspy.A_functions_base.mempy import \
    calc_mem_col, \
    calc_mem_chi, \
    calc_symm_elem_points_by_index_points, \
    get_uniform_density_col, \
    renormailize_density_col, \
    save_spin_density_into_file,\
    form_basins,\
    calc_point_susceptibility, \
    get_uniform_density_chi,\
    renormailize_density_chi, \
    calc_model_value_by_precalculated_data, \
    calc_chi_atoms

from cryspy.A_functions_base.unit_cell import \
    calc_volume_uc_by_unit_cell_parameters, \
    calc_sthovl_by_unit_cell_parameters, \
    calc_eq_ccs_by_unit_cell_parameters
    
from cryspy.A_functions_base.structure_factor import \
    calc_f_nucl_by_dictionary
from cryspy.A_functions_base.flip_ratio import \
    calc_iint, calc_flip_ratio_by_iint, \
    calc_asymmetry_by_iint
from cryspy.A_functions_base.extinction import \
    calc_extinction_sphere
from cryspy.A_functions_base.orbital_functions import \
    calc_density_spherical

from cryspy.A_functions_base.matrix_operations import \
    calc_vv_as_v1_v2_v1

from cryspy.A_functions_base.function_1_error_simplex import \
    error_estimation_simplex


def mempy_reconstruction_by_dictionary(dict_crystal, dict_mem_parameters, l_dict_diffrn, dict_in_out,
        parameter_lambda:float=1.e-5, iteration_max:int=1000, parameter_lambda_min:float=1.e-9, delta_density:float=1.e-5):
    # **Input information about mem parameters**
    
    print("*******************************************")
    print("MEM reconstruction by CrysPy (module MEMPy)")
    print("*******************************************\n")

    print("MEM iteration parameters")
    print("------------------------")
    print(f" starting lambda parameter:     {parameter_lambda*1e6:.3f}*10^-6")
    print(f" maximal number of iterations:  {iteration_max:}")
    print(f" minimal lambda parameter:      {parameter_lambda_min*1e6:}*10^-6")
    print(f" delta_density:                 {delta_density*1e5:}*10^-5\n")

    dict_in_out_keys = dict_in_out.keys()
    print("Density reconstruction")
    print("----------------------")

    n_abc = dict_mem_parameters["points_abc"]
    print(f"Unit cell is devided on points {n_abc[0]:} x {n_abc[1]:} x {n_abc[2]:}.")
    channel_plus_minus = dict_mem_parameters["channel_plus_minus"]
    channel_chi = dict_mem_parameters["channel_chi"]

    if channel_plus_minus:
        magnetization_plus = dict_mem_parameters["magnetization_plus"]
        magnetization_minus = dict_mem_parameters["magnetization_minus"]
        file_spin_density = dict_mem_parameters["file_spin_density"]
        dict_in_out["magnetization_plus"] = magnetization_plus
        dict_in_out["magnetization_minus"] = magnetization_minus

    if channel_chi:
        flag_uniform_prior_density = dict_mem_parameters["flag_uniform_prior_density"]
        flag_only_magnetic_basins =  dict_mem_parameters["flag_only_magnetic_basins"]
        file_magnetization_density = dict_mem_parameters["file_magnetization_density"]

    flag_asymmetry =  dict_mem_parameters["flag_asymmetry"]
    gof_desired = dict_mem_parameters["gof_desired"]

    # **Input information about crystal**
    unit_cell_parameters = dict_crystal["unit_cell_parameters"]
    full_symm_elems = dict_crystal["full_symm_elems"]

    volume_unit_cell = calc_volume_uc_by_unit_cell_parameters(unit_cell_parameters, flag_unit_cell_parameters=False)[0]

    reduced_symm_elems = dict_crystal["reduced_symm_elems"]
    centrosymmetry = dict_crystal["centrosymmetry"]
    if centrosymmetry:
        centrosymmetry_position = dict_crystal["centrosymmetry_position"]
    else:
        centrosymmetry_position = None
    translation_elems = dict_crystal["translation_elems"]

    atom_label = dict_crystal["atom_label"]
    atom_fract_xyz = dict_crystal["atom_fract_xyz"]
    atom_multiplicity = dict_crystal["atom_multiplicity"]

    if channel_chi:
        atom_para_label = dict_crystal["atom_para_label"]
        atom_para_susceptibility = dict_crystal["atom_para_susceptibility"]
        atom_para_sc_chi = dict_crystal["atom_para_sc_chi"]

    # **Index in asymmetric unit cell**
    print("Calculation of asymmetric unit cell...", end="\r")
    index_auc, point_multiplicity = calc_asymmetric_unit_cell_indexes(n_abc, full_symm_elems)
    symm_elem_auc = calc_symm_elem_points_by_index_points(index_auc, n_abc)
    print(f"Number of points in asymmetric unit cell is {index_auc.shape[1]:}.", end="\n")

    # **Basin devision**
    if channel_chi and flag_only_magnetic_basins:
        print("Devision of asymmetric unit cell on bassins...", end="\r")
        flag_atom_para = numpy.any(numpy.expand_dims(atom_label, axis=1) == numpy.expand_dims(atom_para_label, axis=0), axis=1)

        flag_chi, atom_label_auc_chi, atom_multiplicity_auc_chi, atom_distance_auc_chi, atom_symm_elems_auc_chi = \
            form_basins(symm_elem_auc, full_symm_elems, unit_cell_parameters, atom_label[flag_atom_para],
                        atom_fract_xyz[:,flag_atom_para], atom_multiplicity[flag_atom_para], atom_para_label)
        dict_in_out["atom_multiplicity_channel_chi"] = atom_multiplicity_auc_chi
        print(f"Magnetic basins occupy entire unit cell.           \n(flag_only_magnetic_basins: {flag_only_magnetic_basins:})\n")
    elif channel_chi:
        print("Devision of asymmetric unit cell on bassins...", end="\r")
        flag_chi, atom_label_auc_chi, atom_multiplicity_auc_chi, atom_distance_auc_chi, atom_symm_elems_auc_chi = \
            form_basins(symm_elem_auc, full_symm_elems, unit_cell_parameters, atom_label,
                        atom_fract_xyz, atom_multiplicity, atom_para_label)   
        dict_in_out["atom_multiplicity_channel_chi"] = atom_multiplicity_auc_chi
        print(f"Magnetic basins occupy area around magnetic atoms.          \n(flag_only_magnetic_basins: {flag_only_magnetic_basins:})\n")

    if channel_chi:
        index_auc_chi = index_auc[:, flag_chi]
        point_multiplicity_chi = point_multiplicity[flag_chi]
        dict_in_out["point_multiplicity_channel_chi"] = point_multiplicity_chi
        symm_elem_auc_chi = symm_elem_auc[:, flag_chi]
        dict_in_out["symm_elem_channel_chi"] = symm_elem_auc_chi

    if channel_plus_minus and channel_chi:
        flag_col = numpy.logical_not(flag_chi)

        index_auc_col = index_auc[:, flag_col]
        point_multiplicity_col = point_multiplicity[flag_col]
        symm_elem_auc_col = symm_elem_auc[:, flag_col]
        dict_in_out["point_multiplicity_channel_plus_minus"] = point_multiplicity_col
        dict_in_out["symm_elem_channel_plus_minus"] = symm_elem_auc_col

    elif channel_plus_minus:
        index_auc_col = numpy.copy(index_auc)
        point_multiplicity_col = numpy.copy(point_multiplicity)
        symm_elem_auc_col = numpy.copy(symm_elem_auc)
        dict_in_out["point_multiplicity_channel_plus_minus"] = point_multiplicity_col
        dict_in_out["symm_elem_channel_plus_minus"] = symm_elem_auc_col

    print(f"channel_plus_minus: {channel_plus_minus:}")
    print(f"channel_chi:        {channel_chi:}\n")

    if channel_plus_minus:
        print(f"Magnetization of unit cell: {magnetization_plus+magnetization_minus:.3f}  mu_B")
        print(f"(positive channel {magnetization_plus:.3f} mu_B, negative channel {magnetization_minus:.3f}  mu_B)")
        print(f"\nNumber of density points for channel_plus_minus is {index_auc_col.shape[1]}.")
    if channel_chi:
        print(f"Number of density points for channel_chi is {index_auc_chi.shape[1]}.")

    # **Susceptibility tensor $(3\times 3)$ for each point in magnetic basin**
    if channel_chi:
        print("Calculation of restriction on susceptibility...", end="\r")
        point_susceptibility = calc_point_susceptibility(
            unit_cell_parameters, atom_symm_elems_auc_chi, atom_label_auc_chi,
            atom_para_label, atom_para_susceptibility, atom_para_sc_chi, full_symm_elems, symm_elem_auc_chi)
        dict_in_out["susceptibility_channel_chi"] = point_susceptibility
        print(80*" ", end="\r")

    # **Prior density**
    number_unit_cell = numpy.prod(n_abc)
    print("\nCalculation of prior density...         ", end="\r")
    if channel_chi:
        if flag_uniform_prior_density:
            density_chi_prior = get_uniform_density_chi(point_multiplicity_chi, atom_label_auc_chi, atom_multiplicity_auc_chi, volume_unit_cell, number_unit_cell)
            print("Prior density in channel chi is uniform.        ")
        else:
            density_chi_prior = numpy.zeros_like(atom_distance_auc_chi)
            for label in atom_para_label:
                flag_atom = atom_label_auc_chi==label

                dict_shell = dict_crystal[f"shell_{label:}"]
                kappa = float(dict_crystal["mag_atom_kappa"][dict_crystal["mag_atom_label"] == label])
                den_atom = calc_density_spherical(
                    atom_distance_auc_chi[flag_atom], dict_shell["core_population"], dict_shell["core_coeff"], dict_shell["core_zeta"],
                    dict_shell["core_n"], kappa)
                density_chi_prior[flag_atom] = den_atom
            density_chi_prior = renormailize_density_chi(density_chi_prior, point_multiplicity_chi, atom_label_auc_chi, atom_multiplicity_auc_chi, volume_unit_cell, number_unit_cell)
            print("Prior density in channel chi is core.            ")
    if channel_plus_minus:
        density_col_prior = get_uniform_density_col(point_multiplicity_col, volume_unit_cell, number_unit_cell)
        print("Prior density in channel plus-minus is uniform.          ")

    # **Input information about experiments**
    flag_use_precalculated_data = False
    l_exp_value_sigma = []
    l_mem_chi, l_mem_col = [], []
    print(f"Number of experiments is {len(l_dict_diffrn):}.           ")
    for dict_diffrn in l_dict_diffrn:
        if "dict_in_out_"+dict_diffrn["type_name"] in dict_in_out_keys:
            diffrn_dict_in_out = dict_in_out["dict_in_out_"+dict_diffrn["type_name"]]
        else:
            diffrn_dict_in_out = {}
            dict_in_out["dict_in_out_"+dict_diffrn["type_name"]] = diffrn_dict_in_out

        index_hkl = dict_diffrn["index_hkl"]
        h_ccs = dict_diffrn["magnetic_field"]    
        eh_ccs = dict_diffrn["matrix_u"][6:]
        print(f"Preliminary calculation for experiment {dict_diffrn['name']:}...", end="\r")

        diffrn_dict_in_out["index_hkl"] = index_hkl
        diffrn_dict_in_out_keys = diffrn_dict_in_out.keys()
        if channel_plus_minus:
            if "dict_in_out_col" in diffrn_dict_in_out_keys:
                dict_in_out_col = diffrn_dict_in_out["dict_in_out_col"]
            else:
                dict_in_out_col = {}
                diffrn_dict_in_out["dict_in_out_col"] = dict_in_out_col
            
            mem_col = calc_mem_col(
                index_hkl, unit_cell_parameters, eh_ccs, full_symm_elems, symm_elem_auc_col, 
                volume_unit_cell, number_unit_cell, 
                point_multiplicity=point_multiplicity_col,
                dict_in_out=dict_in_out_col, flag_use_precalculated_data=flag_use_precalculated_data)
            
            diffrn_dict_in_out["mem_col"] = mem_col
            l_mem_col.append(mem_col)

        if channel_chi:
            if "dict_in_out_chi" in diffrn_dict_in_out_keys:
                dict_in_out_chi = diffrn_dict_in_out["dict_in_out_chi"]
            else:
                dict_in_out_chi = {}
                diffrn_dict_in_out["dict_in_out_chi"] = dict_in_out_chi
            mem_chi = calc_mem_chi(
                index_hkl, unit_cell_parameters, h_ccs, full_symm_elems, symm_elem_auc_chi,
                point_susceptibility, volume_unit_cell, number_unit_cell, 
                point_multiplicity=point_multiplicity_chi,
                dict_in_out=dict_in_out_chi, flag_use_precalculated_data=flag_use_precalculated_data)
            diffrn_dict_in_out["mem_chi"] = mem_chi
            l_mem_chi.append(mem_chi)

        f_nucl, dder = calc_f_nucl_by_dictionary(
        dict_crystal, diffrn_dict_in_out, flag_use_precalculated_data=flag_use_precalculated_data)
        diffrn_dict_in_out["f_nucl"] = f_nucl


        flip_ratio_es  = dict_diffrn["flip_ratio_es"]
        if flag_asymmetry:
            asymmetry_e = (flip_ratio_es[0] -1.)/(flip_ratio_es[0] + 1.)
            asymmetry_s = numpy.sqrt(2.)*flip_ratio_es[1] * numpy.sqrt(numpy.square(flip_ratio_es[0]) + 1.)/numpy.square(flip_ratio_es[0] + 1.)
            asymmetry_es = numpy.stack([asymmetry_e, asymmetry_s], axis=0)
            l_exp_value_sigma.append(asymmetry_es)
        else:
            l_exp_value_sigma.append(flip_ratio_es)

        

    exp_value_sigma = numpy.concatenate(l_exp_value_sigma, axis=1)    
    if channel_plus_minus:
        mem_col = numpy.concatenate(l_mem_col, axis=1)
    if channel_chi:
        mem_chi = numpy.concatenate(l_mem_chi, axis=1)

    print(f"Total number of reflections is {exp_value_sigma.shape[1]: }.              ")
    if flag_asymmetry:
        print("Density reconstruction is based on asymmetry parameters.")
    else:
        print("Density reconstruction is based on flip ratios.         ")

    # **Preaparation to MEM itertion procedure**
    if channel_plus_minus:
        density_col = numpy.copy(density_col_prior)
        density_col_next = numpy.copy(density_col_prior)
    if channel_chi:
        density_chi = numpy.copy(density_chi_prior)
        density_chi_next = numpy.copy(density_chi_prior)

    # **MEM iteration**
    print("\nMEM iteration procedure")
    print("-----------------------")
    print(f"Desired GoF is {gof_desired:.2f}.")
    c_desired = gof_desired

    c_previous = numpy.inf
    if channel_plus_minus:
        der_c_den_col_previous = numpy.zeros_like(density_col_prior)
    if channel_chi:
        der_c_den_chi_previous = numpy.zeros_like(density_chi_prior)
    iteration = 0
    flag_next = True
    while flag_next:
        iteration += 1

        if channel_plus_minus:
            density_col = numpy.copy(density_col_next)
        if channel_chi:
            density_chi = numpy.copy(density_chi_next)
        l_model_value = []
        l_der_model_den_pm, l_der_model_den_chi = [], []
        for dict_diffrn in l_dict_diffrn:
            diffrn_dict_in_out = dict_in_out["dict_in_out_"+dict_diffrn['type_name']]
            index_hkl = diffrn_dict_in_out["index_hkl"]

            f_m_perp = numpy.zeros(index_hkl.shape, dtype=complex)
            if channel_plus_minus:
                mem_col_exp = diffrn_dict_in_out["mem_col"]
                hh = numpy.expand_dims(numpy.expand_dims(magnetization_plus * density_col[0] + magnetization_minus * density_col[1], axis=0), axis=1)
                f_m_perp_col = (hh*mem_col_exp).sum(axis=2)
                f_m_perp += f_m_perp_col

            if channel_chi:
                mem_chi_exp = diffrn_dict_in_out["mem_chi"]
                f_m_perp_chi = (density_chi*mem_chi_exp).sum(axis=2)
                f_m_perp += f_m_perp_chi


            beam_polarization = dict_diffrn["beam_polarization"]
            flipper_efficiency = dict_diffrn["flipper_efficiency"]
            matrix_u = dict_diffrn["matrix_u"]
            flip_ratio_es = dict_diffrn["flip_ratio_es"]

            f_nucl = diffrn_dict_in_out["f_nucl"]

            wavelength = dict_diffrn["wavelength"]
            sthovl = calc_sthovl_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=False)[0]
            cos_2theta = numpy.cos(2*numpy.arcsin(sthovl*wavelength))

            extinction_model = dict_diffrn["extinction_model"]
            extinction_radius = dict_diffrn["extinction_radius"]
            extinction_mosaicity = dict_diffrn["extinction_mosaicity"]

            func_extinction = lambda f_sq, flag_f_sq: calc_extinction_sphere(
                    f_sq, extinction_radius, extinction_mosaicity, volume_unit_cell, cos_2theta, wavelength,
                    extinction_model, flag_f_sq=False, flag_radius=False,
                    flag_mosaicity=False,
                    flag_volume_unit_cell=False,
                    flag_cos_2theta=False,
                    flag_wavelength=False)

            axis_z = matrix_u[6:9] 
            iint_plus, iint_minus, dder_plus, dder_minus = calc_iint(
                beam_polarization, flipper_efficiency, f_nucl, f_m_perp, axis_z, func_extinction = func_extinction,
                flag_beam_polarization = False, flag_flipper_efficiency = False,
                flag_f_nucl = False, flag_f_m_perp = True,
                dict_in_out = dict_in_out)

            diffrn_dict_in_out["flip_ratio"] = iint_plus/iint_minus

            der_int_plus_fm_perp_real = dder_plus["f_m_perp_real"]
            der_int_plus_fm_perp_imag = dder_plus["f_m_perp_imag"]
            der_int_minus_fm_perp_real = dder_minus["f_m_perp_real"]
            der_int_minus_fm_perp_imag = dder_minus["f_m_perp_imag"]

            if flag_asymmetry:
                model_exp, dder_model_exp = calc_asymmetry_by_iint(
                    iint_plus, iint_minus, c_lambda2=None, iint_2hkl=None,
                    flag_iint_plus=True, flag_iint_minus=True, 
                    flag_c_lambda2=False, flag_iint_2hkl=False)
            else:
                model_exp, dder_model_exp = calc_flip_ratio_by_iint(
                    iint_plus, iint_minus, c_lambda2=None, iint_2hkl=None,
                    flag_iint_plus=True, flag_iint_minus=True, 
                    flag_c_lambda2=False, flag_iint_2hkl=False)

            l_model_value.append(model_exp)
            der_model_int_plus = numpy.expand_dims(dder_model_exp["iint_plus"], axis=0)
            der_model_int_minus = numpy.expand_dims(dder_model_exp["iint_minus"], axis=0)

            if channel_plus_minus:
                der_model_den_pm_exp = (
                    (mem_col_exp.real*numpy.expand_dims(
                    der_model_int_plus*der_int_plus_fm_perp_real +
                    der_model_int_minus*der_int_minus_fm_perp_real, axis=2)
                    ).sum(axis=0) +
                    (mem_col_exp.imag*numpy.expand_dims(
                    der_model_int_plus*der_int_plus_fm_perp_imag +
                    der_model_int_minus*der_int_minus_fm_perp_imag, axis=2)
                    ).sum(axis=0))
                l_der_model_den_pm.append(der_model_den_pm_exp)

            if channel_chi:
                der_model_den_chi_exp = (
                    (mem_chi_exp.real*numpy.expand_dims(
                    der_model_int_plus*der_int_plus_fm_perp_real +
                    der_model_int_minus*der_int_minus_fm_perp_real, axis=2)
                    ).sum(axis=0) +
                    (mem_chi_exp.imag*numpy.expand_dims(
                    der_model_int_plus*der_int_plus_fm_perp_imag +
                    der_model_int_minus*der_int_minus_fm_perp_imag, axis=2)
                    ).sum(axis=0))
                l_der_model_den_chi.append(der_model_den_chi_exp)

        model_value = numpy.concatenate(l_model_value, axis=0)

        diff_value = (exp_value_sigma[0]-model_value)/exp_value_sigma[1]
        c = numpy.square(diff_value).sum(axis=0)/diff_value.shape[0]

        if channel_plus_minus:
            der_model_den_pm = numpy.concatenate(l_der_model_den_pm, axis=0)
            der_c_den_pm = (-2.)/diff_value.shape[0] * (
                numpy.expand_dims((diff_value/exp_value_sigma[1]),axis=1) *
                der_model_den_pm).sum(axis=0)

            der_c_den_col = numpy.stack([magnetization_plus * der_c_den_pm, magnetization_minus * der_c_den_pm], axis=0)

        if channel_chi:
            der_model_den_chi = numpy.concatenate(l_der_model_den_chi, axis=0)
            der_c_den_chi = (-2.)/diff_value.shape[0] * (
                numpy.expand_dims((diff_value/exp_value_sigma[1]),axis=1) *
                der_model_den_chi).sum(axis=0)

        if c > c_previous:
            parameter_lambda = 0.5 * parameter_lambda
            c = c_previous
            if channel_plus_minus:
                density_col = numpy.copy(density_col_previous)
                der_c_den_col = der_c_den_col_previous
            if channel_chi:
                density_chi = numpy.copy(density_chi_previous)
                der_c_den_chi = der_c_den_chi_previous
        else:
            c_previous = c
            parameter_lambda = 1.03 * parameter_lambda
            if channel_plus_minus:
                density_col_previous = numpy.copy(density_col)
                der_c_den_col_previous = der_c_den_col
            if channel_chi:
                density_chi_previous = numpy.copy(density_chi)
                der_c_den_chi_previous = der_c_den_chi

            print(f"Iteration {iteration:5}, lambda {parameter_lambda*1e6:.3f}*10^-6, chi_sq: {c:.2f}              ", end='\r')

        if channel_plus_minus:
            coeff = (parameter_lambda*number_unit_cell/(c_desired*volume_unit_cell))/point_multiplicity_col
            hh = (density_col+delta_density)*numpy.exp(-coeff*der_c_den_col)-delta_density
            hh = numpy.where(hh>0, hh, 0)
            density_col_next = renormailize_density_col(hh, point_multiplicity_col, volume_unit_cell, number_unit_cell)

        if channel_chi:
            coeff = (parameter_lambda*number_unit_cell/(c_desired*volume_unit_cell))*atom_multiplicity_auc_chi/point_multiplicity_chi
            hh = (density_chi+delta_density)*numpy.exp(-coeff*der_c_den_chi)-delta_density
            hh = numpy.where(hh>0, hh, 0)
            density_chi_next = renormailize_density_chi(hh, point_multiplicity_chi, atom_label_auc_chi, atom_multiplicity_auc_chi, volume_unit_cell, number_unit_cell)


        if iteration >= iteration_max:
            flag_next = False
            print(f"Maximal number of iteration is reached ({iteration:}).         ", end='\n')
        if parameter_lambda < parameter_lambda_min:
            flag_next = False
            print(f"Minimal value of parameter lambda {parameter_lambda*1e6:.3f}*10^-6 is reached at iteration {iteration:}.            ", end='\n')
        if c <= c_desired:
            flag_next = False
            print(f"Desired value is reached at iteration {iteration:}.               ", end='\n')

    c_best = c_previous
    print(f"Chi_sq best is {c_best:.2f}")
    if channel_plus_minus:
        density_col_best = numpy.copy(density_col_previous)
        dict_in_out["density_channel_plus_minus"] = density_col_best
    if channel_chi:
        density_chi_best = numpy.copy(density_chi_previous)
        dict_in_out["density_channel_chi"] = density_chi

    # **Save to .den file**
    if channel_plus_minus and (file_spin_density is not None):
        spin_density = density_col_best * numpy.array([[magnetization_plus, ], [magnetization_minus, ]], dtype=float)
        save_spin_density_into_file(file_spin_density, index_auc_col, spin_density, n_abc, unit_cell_parameters,
                reduced_symm_elems, translation_elems, centrosymmetry, centrosymmetry_position)
        print(f"\nReconstructed spin density is written in file '{file_spin_density:}'.")
    if channel_chi and (file_magnetization_density is not None):
        spin_density = numpy.stack([density_chi_best, numpy.zeros_like(density_chi_best)], axis=0)
        save_spin_density_into_file(file_magnetization_density, index_auc_chi, spin_density, n_abc, unit_cell_parameters,
            reduced_symm_elems, translation_elems, centrosymmetry, centrosymmetry_position)
        print(f"\nReconstructed magnetization density is written in file '{file_magnetization_density:}'.")


def mempy_susceptibility_refinement(dict_channel_chi, dict_crystal, dict_mem_parameters, l_dict_diffrn, dict_in_out):
    print("****************************************")
    print("Susceptibility refinement (module MEMPy)")
    print("****************************************")

    number_points = numpy.prod(dict_mem_parameters["points_abc"])
    
    flag_asymmetry =  dict_mem_parameters["flag_asymmetry"]
    channel_plus_minus = dict_mem_parameters["channel_plus_minus"]
    channel_chi = dict_mem_parameters["channel_chi"]
    print(f"Channel plus/minus is {channel_plus_minus:}")
    print("ATTENTION: Channel plus/minus is not taken into account.")
    print(f"Channel chi is {channel_chi:}")

    print(f"Flag asymmetry is {flag_asymmetry:}")
    
    if channel_plus_minus:
        magnetization_plus = dict_mem_parameters["magnetization_plus"]
        magnetization_minus = dict_mem_parameters["magnetization_minus"]
    
    symm_elem_channel_chi = dict_channel_chi["symm_elem_channel_chi"]
    atom_multiplicity_channel_chi = dict_channel_chi["atom_multiplicity_channel_chi"]
    density_channel_chi = dict_channel_chi["density_channel_chi"]
    point_multiplicity_channel_chi = dict_channel_chi["point_multiplicity_channel_chi"]
    
    unit_cell_parameters = dict_crystal["unit_cell_parameters"]
    full_symm_elems = dict_crystal["full_symm_elems"]
    atom_fract_xyz = dict_crystal["atom_fract_xyz"]
    atom_para_sc_chi = dict_crystal["atom_para_sc_chi"]
    atom_para_index = dict_crystal["atom_para_index"]
    atom_para_label = dict_crystal["atom_para_label"]
    atom_para_susceptibility = dict_crystal["atom_para_susceptibility"]
    flags_atom_para_susceptibility = dict_crystal["flags_atom_para_susceptibility"]
    print(f"Number of refined parameters is {flags_atom_para_susceptibility.sum():}.")
    if flags_atom_para_susceptibility.sum() == 0:
        print("There is no refined susceptibility parameters.")
        return
    
    atom_para_fract_xyz = atom_fract_xyz[:, atom_para_index]
    
    n_atom_para = atom_para_susceptibility.shape[1]
    
    print("Preliminary calculations of chi atoms ...", end="\r")
    
    l_exp_value_sigma = []
    for dict_diffrn in l_dict_diffrn:
        flag_use_precalculated_data = False
        index_hkl = dict_diffrn["index_hkl"] 
        diffrn_dict_in_out = {"index_hkl": index_hkl}
    
        chi_atoms = calc_chi_atoms(
            unit_cell_parameters, number_points, full_symm_elems,
            index_hkl, atom_para_fract_xyz, atom_para_sc_chi, 
            symm_elem_channel_chi, point_multiplicity_channel_chi, density_channel_chi)
    
        diffrn_dict_in_out["chi_atoms"] = chi_atoms
    
        eq_ccs, dder = calc_eq_ccs_by_unit_cell_parameters(index_hkl, unit_cell_parameters)
        vp, dder = calc_vv_as_v1_v2_v1(eq_ccs)
        diffrn_dict_in_out["vp"] = vp
    
        f_nucl, dder = calc_f_nucl_by_dictionary(
            dict_crystal, diffrn_dict_in_out, flag_use_precalculated_data=flag_use_precalculated_data)
        diffrn_dict_in_out["f_nucl"] = f_nucl

        dict_in_out["dict_in_out_"+dict_diffrn['type_name']] = diffrn_dict_in_out
    
        flip_ratio_es  = dict_diffrn["flip_ratio_es"]

        if flag_asymmetry:
            asymmetry_e = (flip_ratio_es[0] -1.)/(flip_ratio_es[0] + 1.)
            asymmetry_s = numpy.sqrt(2.)*flip_ratio_es[1] * numpy.sqrt(numpy.square(flip_ratio_es[0]) + 1.)/numpy.square(flip_ratio_es[0] + 1.)
            asymmetry_es = numpy.stack([asymmetry_e, asymmetry_s], axis=0)
            l_exp_value_sigma.append(asymmetry_es)
        else:
            l_exp_value_sigma.append(flip_ratio_es)
    exp_value_sigma = numpy.concatenate(l_exp_value_sigma, axis=1)


    
    def calc_chi_sq(param):
        atom_para_susceptibility[flags_atom_para_susceptibility] = param
        model_value = calc_model_value_by_precalculated_data(atom_para_susceptibility, unit_cell_parameters, flag_asymmetry, dict_in_out, l_dict_diffrn)
    
        chi_sq = numpy.square((model_value-exp_value_sigma[0])/exp_value_sigma[1]).sum()
        return chi_sq
    
    param_0 = atom_para_susceptibility[flags_atom_para_susceptibility]
    chi_sq_per_n = calc_chi_sq(param_0)/exp_value_sigma.shape[1]
    print(70*" ")
    print("Before susceptibility refinement")
    print("Susceptibility tensor:")
    for ind_at, label in enumerate(atom_para_label):
        print(f"{label:5}  {atom_para_susceptibility[0, ind_at]:.5f} {atom_para_susceptibility[1, ind_at]:.5f} {atom_para_susceptibility[2, ind_at]:.5f} {atom_para_susceptibility[3, ind_at]:.5f} {atom_para_susceptibility[4, ind_at]:.5f} {atom_para_susceptibility[5, ind_at]:.5f}")
        
    print(f"chi_sq_per_n is {chi_sq_per_n:.2f}.")
    print("Minimization procedure ...", end="\r")
    res = scipy.optimize.minimize(calc_chi_sq, param_0, method="Nelder-Mead")
    apss = None
    if "hess_inv" in res.keys():
        hess_inv = res["hess_inv"]
        dict_in_out["hess_inv"] = hess_inv
        sigma_p = numpy.sqrt(numpy.abs(numpy.diag(hess_inv)))
        atom_para_susceptibility_sigma = numpy.zeros_like(atom_para_susceptibility)
        atom_para_susceptibility_sigma[flags_atom_para_susceptibility] = sigma_p
        apss = (atom_para_sc_chi * numpy.expand_dims(atom_para_susceptibility_sigma, axis=0)).sum(axis=1)
        dict_in_out["atom_para_susceptibility_sigma"] = apss
    elif "final_simplex" in res.keys():
        n = exp_value_sigma.shape[1]
        m_error, dist_hh = error_estimation_simplex(
            res["final_simplex"][0], res["final_simplex"][1], calc_chi_sq)
        l_sigma = []
        for i, val_2 in zip(range(m_error.shape[0]), dist_hh):
            # slightly change definition, instead of (n-k) here is n
            error = (abs(m_error[i, i])*1./n)**0.5
            if m_error[i, i] < 0.:
                pass
                # warn("Negative diagonal elements of Hessian.", UserWarning)
            if val_2 > error:
                pass
                # warn("Minimum is not found.", UserWarning)
            l_sigma.append(max(error, val_2))
        sigma_p = numpy.array(l_sigma)
        atom_para_susceptibility_sigma = numpy.zeros_like(atom_para_susceptibility)
        atom_para_susceptibility_sigma[flags_atom_para_susceptibility] = sigma_p
        apss = (atom_para_sc_chi * numpy.expand_dims(atom_para_susceptibility_sigma, axis=0)).sum(axis=1)
        dict_in_out["atom_para_susceptibility_sigma"] = apss
        print(sigma_p)
    print(70*" ")
    chi_sq_per_n = calc_chi_sq(res.x)/exp_value_sigma.shape[1]
    
    atom_para_susceptibility[flags_atom_para_susceptibility] = res.x
    atom_para_susceptibility = (atom_para_sc_chi * numpy.expand_dims(atom_para_susceptibility, axis=0)).sum(axis=1)     
    dict_crystal["atom_para_susceptibility"] = atom_para_susceptibility
    print("After susceptibility refinement")
    print("Susceptibility tensor:")
    for ind_at, label in enumerate(atom_para_label):
        print(f"{label:5}  {atom_para_susceptibility[0, ind_at]:8.5f} {atom_para_susceptibility[1, ind_at]:8.5f} {atom_para_susceptibility[2, ind_at]:8.5f} {atom_para_susceptibility[3, ind_at]:8.5f} {atom_para_susceptibility[4, ind_at]:8.5f} {atom_para_susceptibility[5, ind_at]:8.5f}")
        if apss is not None:
            print(f"sigma  {apss[0, ind_at]:8.5f} {apss[1, ind_at]:8.5f} {apss[2, ind_at]:8.5f} {apss[3, ind_at]:8.5f} {apss[4, ind_at]:8.5f} {apss[5, ind_at]:8.5f}")

    print(f"chi_sq_per_n is {chi_sq_per_n:.2f}.")
    
    print(70*"*")
    print("End of MEMPy procedure for susceptibility refinement")
    print(70*"*")
    return 


def mempy_cycle_density_susceptibility(dict_crystal, dict_mem_parameters, l_dict_diffrn, dict_in_out,
        parameter_lambda:float=1.e-5, iteration_max:int=1000, parameter_lambda_min:float=1.e-9, delta_density:float=1.e-5, n_cycle:int=10):
    print(70*"*")
    print("MEMPy: cycle iteration")
    print(70*"*")
    print(f"Number of cycles is {n_cycle:}")
    print(70*" ")
    for i_cycle in range(n_cycle):
        print(f"Cycle {i_cycle+1:}")
        print(len(f"Cycle {i_cycle+1:}")*"-")

        dict_in_out_den = {}
        mempy_reconstruction_by_dictionary(dict_crystal, dict_mem_parameters, l_dict_diffrn, dict_in_out_den,
            parameter_lambda=parameter_lambda, iteration_max=iteration_max, parameter_lambda_min=parameter_lambda_min, delta_density=delta_density)
    
        dict_channel_chi = {
            'atom_multiplicity_channel_chi': dict_in_out_den['atom_multiplicity_channel_chi'],
            'point_multiplicity_channel_chi': dict_in_out_den['point_multiplicity_channel_chi'],
            'symm_elem_channel_chi': dict_in_out_den['symm_elem_channel_chi'],
            'susceptibility_channel_chi': dict_in_out_den['susceptibility_channel_chi'],
            'density_channel_chi': dict_in_out_den['density_channel_chi'],
        }
    
        dict_in_out_susc = {} 
        mempy_susceptibility_refinement(dict_channel_chi, dict_crystal, dict_mem_parameters, l_dict_diffrn, dict_in_out_susc)
        print(70*" ")
    
    dict_in_out["dict_in_out_den"] = dict_in_out_den
    dict_in_out["dict_in_out_susc"] = dict_in_out_susc
    return 